import dataclasses
from collections.abc import AsyncGenerator

from pymobiledevice3.remote.core_device.core_device_service import CoreDeviceService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


@dataclasses.dataclass
class SysDiagnoseResponse:
    preferred_filename: str
    file_size: int
    generator: AsyncGenerator[bytes, None]


class DiagnosticsServiceService(CoreDeviceService):
    """
    Obtain device diagnostics
    """

    SERVICE_NAME = "com.apple.coredevice.diagnosticsservice"

    def __init__(self, rsd: RemoteServiceDiscoveryService):
        super().__init__(rsd, self.SERVICE_NAME)

    async def capture_sysdiagnose(self, is_dry_run: bool) -> SysDiagnoseResponse:
        response = await self.invoke(
            "com.apple.coredevice.feature.capturesysdiagnose",
            {"options": {"collectFullLogs": True}, "isDryRun": is_dry_run},
        )
        return SysDiagnoseResponse(
            file_size=response["fileTransfer"]["expectedLength"],
            preferred_filename=response["preferredFilename"],
            generator=self.service.iter_file_chunks(response["fileTransfer"]["expectedLength"]),
        )



import logging
from typing import Any, Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError, NoDeviceConnectedError
from pymobiledevice3.lockdown import DEFAULT_LABEL
from pymobiledevice3.restore.restore_options import RestoreOptions
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.usbmux import select_devices_by_connection_type

logger = logging.getLogger(__name__)


class RestoredClient:
    DEFAULT_CLIENT_NAME = "pymobiledevice3"
    SERVICE_PORT = 62078

    @classmethod
    async def create(cls, ecid: str) -> "RestoredClient":
        for mux_device in await select_devices_by_connection_type("USB"):
            logger.debug(f"Iterating: {mux_device}")
            service = await ServiceConnection.create_using_usbmux(
                mux_device.serial, cls.SERVICE_PORT, connection_type=mux_device.connection_type
            )
            await service.start()

            query_type = await service.send_recv_plist({"Request": "QueryType"})
            version = query_type.get("RestoreProtocolVersion")
            logger.debug(f"RestoreProtocolVersion: {version}")

            if query_type.get("Type") != "com.apple.mobile.restored":
                logger.debug(f"Skipping: {mux_device.serial} as its not a restored device")
                await service.close()
                continue

            restored_client = cls(mux_device.serial, version, service)
            await restored_client._connect()

            if restored_client.ecid != ecid:
                logger.debug(
                    f"Skipping: {restored_client.ecid} as its not the right ECID ({restored_client.ecid} "
                    f"instead of {ecid})"
                )
                await service.close()
                continue

            return restored_client
        raise NoDeviceConnectedError()

    def __init__(self, udid: str, version: str, service: ServiceConnection) -> None:
        self.udid = udid
        self.version = version
        self.service = service
        self.label = DEFAULT_LABEL

    async def _connect(self) -> None:
        self.hardware_info = (await self.query_value("HardwareInfo"))["HardwareInfo"]
        self.ecid = self.hardware_info["UniqueChipID"] & 0xFFFFFFFFFFFFFFFF
        self.saved_debug_info = (await self.query_value("SavedDebugInfo"))["SavedDebugInfo"]

    @staticmethod
    async def _get_or_verify_udid(udid: Optional[str] = None) -> str:
        device = await usbmux.select_device(udid)
        if device is None:
            if udid:
                raise ConnectionFailedError()
            else:
                raise NoDeviceConnectedError()
        return device.serial

    async def query_value(self, key: Optional[str] = None) -> Any:
        req = {"Request": "QueryValue", "Label": self.label}

        if key:
            req["QueryKey"] = key

        return await self.service.send_recv_plist(req)

    async def start_restore(self, opts: Optional[RestoreOptions] = None) -> None:
        req = {"Request": "StartRestore", "Label": self.label, "RestoreProtocolVersion": self.version}

        if opts is not None:
            req["RestoreOptions"] = opts.to_dict()

        logger.debug(f"start_restore request: {req}")

        return await self.service.send_plist(req)

    async def reboot(self) -> dict:
        return await self.service.send_recv_plist({"Request": "Reboot", "Label": self.label})

    async def send(self, message: dict) -> None:
        await self.service.send_plist(message)

    async def recv(self) -> dict:
        return await self.service.recv_plist()



import time
from typing import Literal, Optional

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command
from pymobiledevice3.services.power_assertion import PowerAssertionService

cli = InjectingTyper(
    name="power-assertion",
    no_args_is_help=True,
)


@cli.command("power-assertion")
@async_command
async def power_assertion(
    service_provider: ServiceProviderDep,
    assertion_type: Literal["AMDPowerAssertionTypeWirelessSync", "PreventUserIdleSystemSleep", "PreventSystemSleep"],
    name: str,
    timeout: int,
    details: Optional[str] = None,
) -> None:
    """Create a power assertion"""
    async with PowerAssertionService(service_provider).create_power_assertion(assertion_type, name, timeout, details):
        print("> Hit Ctrl+C to exit")
        time.sleep(timeout)



import asyncio
import logging
from abc import abstractmethod
from typing import Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.service_connection import ServiceConnection


class TcpForwarderBase:
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    MAX_FORWARDED_CONNECTIONS = 200

    def __init__(self, src_port: int, listening_event: Optional[asyncio.Event] = None):
        """
        Initialize a new tcp forwarder

        :param src_port: tcp port to listen on
        :param enable_ssl: enable ssl wrapping for the transferred data
        :param listening_event: asyncio event to fire when the listener is ready
        """
        self.logger = logging.getLogger(__name__)
        self.src_port = src_port
        self.server: Optional[asyncio.Server] = None
        self.stopped = asyncio.Event()
        self.listening_event = listening_event
        self._connection_tasks: set[asyncio.Task] = set()

    async def start(self, address="127.0.0.1"):
        """forward each connection from given local machine port to remote device port"""
        self.server = await asyncio.start_server(
            self._handle_server_connection, address, self.src_port, backlog=self.MAX_FORWARDED_CONNECTIONS
        )
        if self.listening_event:
            self.listening_event.set()
        try:
            await self.stopped.wait()
        finally:
            self.logger.info("Closing everything")
            if self.server is not None:
                self.server.close()
                await self.server.wait_closed()
            tasks = list(self._connection_tasks)
            for task in tasks:
                task.cancel()
            if tasks:
                await asyncio.gather(*tasks, return_exceptions=True)

    @abstractmethod
    async def _establish_remote_connection(self) -> ServiceConnection:
        pass

    async def _pipe(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, peer: str, direction: str
    ) -> None:
        while True:
            data = await reader.read(65536)
            if not data:
                return
            writer.write(data)
            await writer.drain()
            self.logger.debug("%s %s bytes=%d", peer, direction, len(data))

    async def _handle_server_connection(self, local_reader: asyncio.StreamReader, local_writer: asyncio.StreamWriter):
        """accept the connection from local machine and attempt to connect at remote"""
        peer = str(local_writer.get_extra_info("peername"))
        task = asyncio.current_task()
        if task is not None:
            self._connection_tasks.add(task)
        remote_connection: Optional[ServiceConnection] = None
        try:
            try:
                remote_connection = await self._establish_remote_connection()
                await remote_connection.start()
            except ConnectionFailedError:
                self.logger.error("failed to connect to remote endpoint")
                return

            assert remote_connection.reader is not None
            assert remote_connection.writer is not None
            self.logger.info("connection established from %s", peer)
            client_to_remote = asyncio.create_task(
                self._pipe(local_reader, remote_connection.writer, peer, "client->remote")
            )
            remote_to_client = asyncio.create_task(
                self._pipe(remote_connection.reader, local_writer, peer, "remote->client")
            )
            done, pending = await asyncio.wait(
                {client_to_remote, remote_to_client},
                return_when=asyncio.FIRST_COMPLETED,
            )
            for p in pending:
                p.cancel()
            if pending:
                await asyncio.gather(*pending, return_exceptions=True)
            for d in done:
                exc = d.exception()
                if exc is not None and not isinstance(exc, ConnectionResetError):
                    self.logger.debug("connection %s ended with %r", peer, exc)
        finally:
            if remote_connection is not None:
                await remote_connection.close()
            local_writer.close()
            await local_writer.wait_closed()
            if remote_connection is not None:
                self.logger.info("connection %s was closed", peer)
            if task is not None:
                self._connection_tasks.discard(task)

    def stop(self):
        """stop forwarding"""
        self.stopped.set()

    @property
    def listening_port(self) -> Optional[int]:
        """Return the bound local port once the server has started listening."""
        if self.server is None or not self.server.sockets:
            return None
        sockname = self.server.sockets[0].getsockname()
        return sockname[1]


class UsbmuxTcpForwarder(TcpForwarderBase):
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    def __init__(
        self,
        serial: str,
        dst_port: int,
        src_port: int,
        listening_event: Optional[asyncio.Event] = None,
        usbmux_connection_type: Optional[str] = None,
        usbmux_address: Optional[str] = None,
    ):
        """
        Initialize a new tcp forwarder

        :param serial: device serial
        :param dst_port: tcp port to connect to each new connection via the supplied lockdown object
        :param src_port: tcp port to listen on
        :param listening_event: asyncio event to fire when the listener is ready
        :param usbmux_connection_type: preferred connection type
        :param usbmux_address: usbmuxd address
        """
        super().__init__(src_port, listening_event)
        self.serial = serial
        self.dst_port = dst_port
        self.usbmux_connection_type = usbmux_connection_type
        self.usbmux_address = usbmux_address

    async def _establish_remote_connection(self) -> ServiceConnection:
        # connect directly using usbmuxd
        mux_device = await usbmux.select_device(
            self.serial, connection_type=self.usbmux_connection_type, usbmux_address=self.usbmux_address
        )
        self.logger.debug("Selected device: %r", mux_device)
        if mux_device is None:
            raise ConnectionFailedError()
        sock = await mux_device.connect(self.dst_port, usbmux_address=self.usbmux_address)
        return ServiceConnection(sock, mux_device=mux_device)


class LockdownTcpForwarder(TcpForwarderBase):
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    def __init__(
        self,
        service_provider: LockdownServiceProvider,
        src_port: int,
        service_name: str,
        listening_event: Optional[asyncio.Event] = None,
    ):
        """
        Initialize a new tcp forwarder

        :param src_port: tcp port to listen on
        :param service_name: service name to connect to
        :param listening_event: asyncio event to fire when the listener is ready
        """
        super().__init__(src_port, listening_event)
        self.service_provider = service_provider
        self.service_name = service_name

    async def _establish_remote_connection(self) -> ServiceConnection:
        return await self.service_provider.start_lockdown_developer_service(self.service_name)



import logging
from textwrap import dedent

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command, print_json
from pymobiledevice3.services.idam import IDAMService

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="idam",
    help=dedent("""\
        Access IDAM (Inter-Device Audio and MIDI) configuration

        For more info refer to:
        <https://www.youtube.com/watch?v=IXmP938brnc>
    """),
    no_args_is_help=True,
)


@cli.command()
@async_command
async def configuration_inquiry(service_provider: ServiceProviderDep) -> None:
    """Inquiry IDAM configuration"""
    async with IDAMService(service_provider) as idam:
        print_json(await idam.configuration_inquiry())


@cli.command()
@async_command
async def enable(service_provider: ServiceProviderDep) -> None:
    """Enable IDAM"""
    async with IDAMService(service_provider) as idam:
        await idam.set_idam_configuration(True)


@cli.command()
@async_command
async def disable(service_provider: ServiceProviderDep) -> None:
    """Disable IDAM"""
    async with IDAMService(service_provider) as idam:
        await idam.set_idam_configuration(False)



import asyncio

import pytest

from pymobiledevice3.lockdown import LockdownClient

LOCKDOWND_SOCKET_SELECT_TIMEOUT = 60


@pytest.mark.asyncio
async def test_lockdown_reconnect(lockdown: LockdownClient) -> None:
    d1 = await lockdown.get_date()

    # add some threshold to make sure lockdownd closed the connection on its end
    await asyncio.sleep(LOCKDOWND_SOCKET_SELECT_TIMEOUT + 5)

    d2 = await lockdown.get_date()

    assert d1 < d2



from collections.abc import AsyncGenerator
from typing import Any

from pymobiledevice3.dtx import DTXService, dtx_method
from pymobiledevice3.dtx_service import DtxService


class EnergyMonitorService(DTXService):
    IDENTIFIER = "com.apple.xcode.debug-gauge-data-providers.Energy"

    @dtx_method("startSamplingForPIDs:", expects_reply=False)
    async def start_sampling_for_pids_(self, pid_list: list[int]) -> None: ...

    @dtx_method("stopSamplingForPIDs:", expects_reply=False)
    async def stop_sampling_for_pids_(self, pid_list: list[int]) -> None: ...

    @dtx_method("sampleAttributes:forPIDs:")
    async def sample_attributes_for_pids_(self, attributes: dict, pid_list: list[int]) -> Any: ...


class EnergyMonitor(DtxService[EnergyMonitorService]):
    def __init__(self, dvt, pid_list: list[int]) -> None:
        super().__init__(dvt)
        self._pid_list = pid_list

    async def __aenter__(self):
        await self.connect()
        # stop monitoring if already monitored
        await self.service.stop_sampling_for_pids_(self._pid_list)

        await self.service.start_sampling_for_pids_(self._pid_list)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.service.stop_sampling_for_pids_(self._pid_list)

    async def __aiter__(self) -> AsyncGenerator[Any, None]:
        while True:
            yield await self._sample_once()

    async def _sample_once(self):
        return await self.service.sample_attributes_for_pids_({}, self._pid_list)



from __future__ import annotations

import logging
import sys
from typing import Any, ClassVar, Generic, Optional, TypeVar

from typing_extensions import Self

from pymobiledevice3.dtx import DTXConnection
from pymobiledevice3.dtx import DTXDynamicService as _DTXDynamicService
from pymobiledevice3.dtx import DTXService as _DTXService
from pymobiledevice3.dtx_service_provider import DtxServiceProvider

if sys.version_info >= (3, 13):
    _SVC_T = TypeVar("_SVC_T", bound=_DTXService, default=_DTXDynamicService)
else:
    _SVC_T = TypeVar("_SVC_T", bound=_DTXService)


class DtxService(Generic[_SVC_T]):
    CHANNEL_IDENTIFIER: ClassVar[Optional[str]] = None
    """Optional raw channel identifier string.

    When set, the channel is opened by this string rather than by the service
    class resolved from ``_SVC_T`` or defaults to :class:`~pymobiledevice3.dtx.DTXDynamicService`.
    This is the highest-priority resolution path.
    """

    _inferred_service_class: ClassVar[Optional[type[_DTXService]]] = None
    """Service class inferred from the ``_SVC_T`` type parameter.

    Populated automatically by :meth:`__init_subclass__` when the subclass is
    defined as ``DtxService[SomeClass]``.  Do not set this manually.
    """

    def __init_subclass__(cls, **kw: Any) -> None:
        super().__init_subclass__(**kw)
        # Look for a direct DtxService[X] base on *this* class only.
        # Using ``base.__origin__ is DtxService`` (PEP 560 / Python 3.7+) is
        # more precise than inspecting arg-count with typing.get_args —
        # it won't accidentally match DtxProxyService[A, B] or unrelated bases.
        for base in getattr(cls, "__orig_bases__", ()):
            if getattr(base, "__origin__", None) is DtxService:
                args = getattr(base, "__args__", ())
                if args and isinstance(args[0], type) and issubclass(args[0], _DTXService):
                    cls._inferred_service_class = args[0]
                break
        # Note: no validation here — DtxProxyService resets _inferred_service_class
        # to None for its own subclasses and performs its own validation.

    def __init__(
        self,
        provider: DtxServiceProvider,
        service: Optional[_SVC_T] = None,
    ) -> None:
        """
        :param provider: Active (or not-yet-connected) :class:`DtxServiceProvider`
            that owns the underlying
            :class:`~pymobiledevice3.dtx.DTXConnection`.
        :param service: Pre-opened :class:`~pymobiledevice3.dtx.DTXService`
            to re-use.  When supplied the instance does **not** manage its
            lifecycle.
        """
        self._provider = provider
        self._service: Optional[_SVC_T] = service
        self.logger = logging.getLogger(self.__module__)

    # ------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------

    @property
    def provider(self) -> DtxServiceProvider:
        """The owning :class:`DtxServiceProvider`."""
        return self._provider

    @property
    def dtx(self) -> DTXConnection:
        """The underlying :class:`~pymobiledevice3.dtx.DTXConnection`.

        Shorthand for ``self.provider.dtx``.
        """
        return self._provider.dtx

    @property
    def service(self) -> _SVC_T:
        """The active :class:`~pymobiledevice3.dtx.DTXService` channel.

        Raises :exc:`RuntimeError` if :meth:`connect` has not been called yet.
        The return type matches the ``_SVC_T`` type parameter or :class:`~pymobiledevice3.dtx.DTXDynamicService`.
        """
        if self._service is None:
            raise RuntimeError("not connected — use `async with` or await connect()")
        return self._service

    # ------------------------------------------------------------------
    # Hook
    # ------------------------------------------------------------------

    async def _acquire_channel(self) -> _SVC_T:
        """Open and return the DTX channel for this service.

        Resolution order:

        1. :attr:`CHANNEL_IDENTIFIER` set **and** ``_SVC_T`` inferred →
           ``open_channel(identifier, cls)`` (typed lookup by explicit id).
        2. :attr:`CHANNEL_IDENTIFIER` set, no inferred class →
           ``open_channel(identifier)`` (registry/dynamic lookup; used by
           :class:`~pymobiledevice3.dtx_proxy_service.DtxProxyService` so the
           sub-service registry wires the proxy correctly).
        3. Only ``_SVC_T`` inferred → ``open_channel(cls)`` (identifier from
           ``cls.IDENTIFIER``).

        Override for non-standard acquisition, for example to wait for a
        remote-initiated proxied channel::

            async def _acquire_channel(self):
                return await self.dtx.wait_for_proxied_service(
                    XCTestDriverInterface, remote=True, timeout=90.0
                )
        """
        if self.CHANNEL_IDENTIFIER is not None:
            if self._inferred_service_class is not None:
                return await self._provider.dtx.open_channel(self.CHANNEL_IDENTIFIER, self._inferred_service_class)
            return await self._provider.dtx.open_channel(self.CHANNEL_IDENTIFIER)
        assert self._inferred_service_class is not None, (
            "Cannot infer service class — specify CHANNEL_IDENTIFIER or provide a concrete "
            "type parameter, e.g. DtxService[MyService]"
        )
        return await self._provider.dtx.open_channel(self._inferred_service_class)

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    async def connect(self) -> None:
        """Ensure the provider is connected and acquire the channel.

        No-op if already connected.  Connects the provider if it has not been
        connected yet (safe to call even when the provider is already live).
        """
        await self._provider.connect()
        if self._inferred_service_class is not None:
            self._provider.dtx.register_service(self._inferred_service_class)
        if self._service is not None:
            return
        self._service = await self._acquire_channel()

    async def close(self) -> None:
        """No-op — channel lifetime is managed by the :class:`DtxServiceProvider`.

        Override if you need to perform teardown on the channel (e.g. sending
        a graceful shutdown message) before the connection is closed.
        """

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(self, *_: Any) -> None:
        await self.close()



"""Integration test for XCUITest execution via the DTX implementation.

Run with:
    pytest tests/services/test_xcuitest.py -s --tunnel '' -v

Configurations (runner + target bundle IDs, optional env and timeout) are
read from ``tests/services/xcuitest.json``
(see ``xcuitest.json.example`` for the expected format).
Each entry in that JSON array becomes a separate parametrized test run.

Override the config file path:
    pytest --xcuitest-config /path/to/config.json ...
"""

from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import Any

import pytest

from pymobiledevice3.services.dvt.testmanaged.xcuitest import (
    TestConfig as _TestConfig,
)
from pymobiledevice3.services.dvt.testmanaged.xcuitest import (
    XCTestCaseResult,
    XCUITestListener,
    XCUITestService,
)


class _RecordingListener(XCUITestListener):
    """Captures lifecycle events for assertion in tests."""

    def __init__(self) -> None:
        self.events: list[str] = []
        self.runner_ready_event = asyncio.Event()

    async def did_begin_executing_test_plan(self) -> None:
        self.events.append("begin_plan")

    async def did_finish_executing_test_plan(self) -> None:
        self.events.append("finish_plan")

    async def test_suite_did_start(self, suite: str, started_at: str) -> None:
        self.events.append(f"suite_start:{suite}")

    async def test_suite_did_finish(
        self,
        suite,
        finished_at,
        run_count,
        failures,
        unexpected,
        test_duration,
        total_duration,
        skipped,
        expected_failures,
        uncaught_exceptions,
    ) -> None:
        self.events.append(f"suite_finish:{suite}")

    async def test_case_did_start(self, test_class: str, method: str) -> None:
        self.events.append(f"case_start:{test_class}/{method}")

    async def test_case_did_finish(self, result: XCTestCaseResult) -> None:
        self.events.append(f"case_finish:{result.test_class}/{result.method}:{result.status}")

    async def test_case_did_fail(self, test_class, method, message, file, line) -> None:
        self.events.append(f"case_fail:{test_class}/{method}")

    async def log_message(self, message: str) -> None:
        self.events.append(f"log:{message[:60]}")

    # --- overrides to signal test readiness for test_xcuitest_runner_ready ---

    async def test_bundle_ready(self):
        self.runner_ready_event.set()
        return await super().test_bundle_ready()

    async def test_bundle_ready_with_capabilities(self, capabilities):
        self.runner_ready_event.set()
        return await super().test_bundle_ready_with_capabilities(capabilities)

    async def test_bundle_ready_with_protocol_version(self, protocol_version, minimum_version):
        self.runner_ready_event.set()
        return await super().test_bundle_ready_with_protocol_version(protocol_version, minimum_version)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


async def _make_cfg(
    xcuitest_service: XCUITestService,
    xcuitest_cfg: dict[str, Any],
) -> _TestConfig:
    cfg = await _TestConfig.create_for(
        xcuitest_service.lockdown,
        xcuitest_cfg["runner_bundle_id"],
        xcuitest_cfg.get("target_bundle_id"),
    )
    cfg.runner_app_env = xcuitest_cfg.get("runner_env") or None
    return cfg


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_xcuitest_runner_ready(
    xcuitest_service: XCUITestService,
    xcuitest_cfg: dict[str, Any],
) -> None:
    """Assert the XCUITest runner reaches the ready state (full handshake verified).

    We do not wait for test-plan completion; verifying
    ``_XCT_testRunnerReadyWithCapabilities_`` fires is sufficient to validate
    the entire DTX connection and proxy setup.
    """
    listener = _RecordingListener()
    cfg = await _make_cfg(xcuitest_service, xcuitest_cfg)
    timeout = xcuitest_cfg.get("timeout", 30.0)

    task = asyncio.create_task(
        xcuitest_service.run(
            cfg,
            timeout=timeout,
            listener=listener,
        )
    )
    try:
        await asyncio.wait_for(listener.runner_ready_event.wait(), timeout=timeout)
    except asyncio.TimeoutError:
        pytest.fail("_XCT_testRunnerReadyWithCapabilities_ was not received within 90s")
    finally:
        task.cancel()
        with suppress(asyncio.CancelledError, TimeoutError, Exception):
            await task

    print(f"\n[listener events] {listener.events}")


@pytest.mark.asyncio
async def test_xcuitest_full_completion(
    xcuitest_service: XCUITestService,
    xcuitest_cfg: dict[str, Any],
) -> None:
    """Assert the test plan runs to completion.

    Key regression test: the runner must receive the XCTestConfiguration reply
    on the correct channel code, load the test bundle, run all tests, and fire
    ``_XCT_didFinishExecutingTestPlan``.
    """
    listener = _RecordingListener()
    cfg = await _make_cfg(xcuitest_service, xcuitest_cfg)
    await xcuitest_service.run(cfg, timeout=xcuitest_cfg.get("timeout", 30.0), listener=listener)
    print(f"\n[listener events] {listener.events}")
    assert "finish_plan" in listener.events, f"Test plan did not finish. Events: {listener.events}"



import asyncio
import pathlib
from datetime import datetime

import pytest
import pytest_asyncio

from pymobiledevice3.exceptions import AfcException, AfcFileNotFoundError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import MAXIMUM_READ_SIZE, AfcError, AfcService

TEST_FILENAME = "test"
TEST_FOLDER_NAME = "test_folder"

pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture(scope="function")
async def afc(lockdown: LockdownClient):
    async with AfcService(lockdown) as afc:
        yield afc


async def test_exists(afc: AfcService) -> None:
    assert await afc.exists("DCIM")
    assert not await afc.exists("a_directory_that_doesnt_exist")


async def test_exists_folder_in_a_file(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"data")
    try:
        assert not await afc.exists(f"{TEST_FILENAME}/sub_folder")
    finally:
        await afc.rm(TEST_FILENAME)


async def test_rm(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"")
    filenames = await afc.listdir("/")
    assert TEST_FILENAME in filenames

    await afc.rm(TEST_FILENAME)

    filenames = await afc.listdir("/")
    assert TEST_FILENAME not in filenames


async def test_rm_force_missing_file(afc: AfcService) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.rm(TEST_FILENAME)
    await afc.rm(TEST_FILENAME, force=True)


@pytest.mark.parametrize(
    "path",
    [
        "file_that_doesnt_exist.txt",
        "missingfolder/file_that_doesnt_exist.txt",
        "/missingfolder/file_that_doesnt_exist.txt",
        "/missingfolder/file_that_doesnt_exist.txt/",
        "/missingfolder/./././file_that_doesnt_exist.txt/",
    ],
)
async def test_rm_file_doesnt_exist(afc: AfcService, path: str) -> None:
    with pytest.raises(AfcFileNotFoundError) as e:
        await afc.rm(path)
    assert e.value.status == AfcError.OBJECT_NOT_FOUND


async def test_get_device_info(afc: AfcService) -> None:
    device_info = await afc.get_device_info()
    assert device_info["Model"].startswith("iPhone")
    assert int(device_info["FSTotalBytes"]) > int(device_info["FSFreeBytes"])


async def test_listdir(afc: AfcService) -> None:
    filenames = await afc.listdir("/")
    assert "DCIM" in filenames
    assert "Downloads" in filenames
    assert "Books" in filenames


@pytest.mark.parametrize(
    "path",
    [
        "missing_folder",
        "missingfolder/missing_folder",
        "missingfolder/missing_folder/",
        "/missingfolder/missing_folder",
        "/missingfolder/missing_folder/",
    ],
)
async def test_listdir_folder_doesnt_exist(afc: AfcService, path: str) -> None:
    with pytest.raises(AfcFileNotFoundError) as e:
        await afc.listdir(path)
    assert e.value.status == AfcError.OBJECT_NOT_FOUND


async def test_listdir_file(afc: AfcService):
    await afc.set_file_contents(TEST_FILENAME, b"data")
    try:
        with pytest.raises(AfcException) as e:
            await afc.listdir(TEST_FILENAME)
    finally:
        await afc.rm(TEST_FILENAME)
    assert e.value.status == AfcError.READ_ERROR


@pytest.mark.parametrize(
    "path",
    [
        "test_dir_a/test_dir_b/test_dir_c/test_dir_d",
        "test_dir_a/test_dir_b/test_dir_c/test_dir_d/",
        "/test_dir_a/test_dir_b/test_dir_c/test_dir_d",
        "/test_dir_a/test_dir_b/test_dir_c/test_dir_d/",
        "/test_dir_a/./../test_dir_a/test_dir_b/../test_dir_b/test_dir_c/",
    ],
)
async def test_makedirs_and_rm_dir(afc: AfcService, path: str) -> None:
    root_dir = pathlib.PosixPath(path.lstrip("/")).parts[0]
    await afc.rm(root_dir, force=True)
    try:
        assert not await afc.exists(path)
        await afc.makedirs(path)
        assert await afc.exists(path)
        await afc.rm(root_dir)
        assert not await afc.exists(path)
    finally:
        await afc.rm(root_dir, force=True)


async def test_makedirs_file(afc: AfcService):
    await afc.set_file_contents(TEST_FILENAME, b"data")
    impossible_path = f"{TEST_FILENAME}/sub_folder"
    try:
        with pytest.raises(AfcException) as e:
            await afc.makedirs(impossible_path)
        assert not await afc.exists(impossible_path)
    finally:
        await afc.rm(TEST_FILENAME)
    assert e.value.status == AfcError.OBJECT_EXISTS


async def test_isdir_file(afc: AfcService):
    await afc.set_file_contents(TEST_FILENAME, b"data")
    assert not await afc.isdir(TEST_FILENAME)
    await afc.rm(TEST_FILENAME)


@pytest.mark.parametrize(
    "path",
    [
        TEST_FOLDER_NAME,
        f"{TEST_FOLDER_NAME}/",
        f"/{TEST_FOLDER_NAME}",
        f"/{TEST_FOLDER_NAME}/",
        f"/{TEST_FOLDER_NAME}/./../{TEST_FOLDER_NAME}/",
    ],
)
async def test_isdir_folder(afc: AfcService, path: str) -> None:
    await afc.makedirs(TEST_FILENAME)
    assert await afc.isdir(TEST_FILENAME)
    await afc.rm(TEST_FILENAME)


async def test_isdir_missing_path(afc: AfcService) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.isdir("folder_that_doesnt_exist")


async def test_isdir_missing_path_inside_a_file(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"data")
    impossible_path = f"{TEST_FILENAME}/sub_folder"
    try:
        with pytest.raises(AfcFileNotFoundError):
            await afc.isdir(impossible_path)
    finally:
        await afc.rm(TEST_FILENAME)


async def test_stat_file(afc: AfcService) -> None:
    data = b"data"
    timestamp = datetime.fromtimestamp(await afc.lockdown.get_value(key="TimeIntervalSince1970"))
    timestamp = timestamp.replace(microsecond=0)  # stat resolution might not include microseconds
    await afc.set_file_contents(TEST_FILENAME, data)
    stat = await afc.stat(TEST_FILENAME)
    await afc.rm(TEST_FILENAME)
    assert stat["st_size"] == len(data)
    assert stat["st_ifmt"] == "S_IFREG"
    assert stat["st_mtime"] >= timestamp


async def test_stat_folder(afc: AfcService) -> None:
    timestamp = datetime.fromtimestamp(await afc.lockdown.get_value(key="TimeIntervalSince1970"))
    timestamp = timestamp.replace(microsecond=0)  # stat resolution might not include microseconds
    await afc.makedirs(TEST_FOLDER_NAME)
    stat = await afc.stat(TEST_FOLDER_NAME)
    await afc.rm(TEST_FOLDER_NAME)
    assert stat["st_size"] in (64, 68)
    assert stat["st_ifmt"] == "S_IFDIR"
    assert stat["st_mtime"] >= timestamp


@pytest.mark.parametrize(
    "path",
    [
        "missing_file",
        "/missing_file",
        "missing_folder/",
        "/missing_folder/",
        "missingfolder/missing_file",
        "missingfolder/missing_folder/",
        "/missingfolder/missing_file",
        "/missingfolder/missing_folder/",
    ],
)
async def test_stat_doesnt_exist(afc: AfcService, path: str) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.stat(path)


async def test_stat_missing_path_inside_a_file(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"data")
    impossible_path = f"{TEST_FILENAME}/sub_folder"
    try:
        with pytest.raises(AfcFileNotFoundError):
            await afc.stat(impossible_path)
    finally:
        await afc.rm(TEST_FILENAME)


async def test_fopen_missing_file(afc: AfcService) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.fopen("file_that_doesnt_exist")


async def test_fclose_not_opened(afc: AfcService) -> None:
    with pytest.raises(AfcException) as e:
        await afc.fclose(77)
    assert e.value.status == AfcError.INVALID_ARG


async def test_rename(afc: AfcService):
    await afc.set_file_contents("source.txt", b"data")
    await afc.rename("source.txt", "dest.txt")
    try:
        assert await afc.get_file_contents("dest.txt") == b"data"
    finally:
        await afc.rm("dest.txt")
    with pytest.raises(AfcFileNotFoundError):
        await afc.get_file_contents("source.txt")


async def test_rename_between_folders(afc: AfcService) -> None:
    await afc.makedirs("dir_a/dir_b")
    source = "dir_a/dir_b/source.txt"
    dest = "dir_a/source.txt"
    await afc.set_file_contents(source, b"data")
    await afc.rename(source, dest)
    try:
        assert await afc.get_file_contents(dest) == b"data"
        with pytest.raises(AfcFileNotFoundError):
            await afc.get_file_contents(source)
    finally:
        await afc.rm("dir_a")


async def test_rename_missing_source(afc: AfcService) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.rename("source.txt", "dest.txt")


async def test_rename_source_path_inside_a_file(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"data")
    impossible_path = f"{TEST_FILENAME}/source.txt"
    try:
        with pytest.raises(AfcFileNotFoundError):
            await afc.rename(impossible_path, "dest.txt")
    finally:
        await afc.rm(TEST_FILENAME)


async def test_rename_dest_path_inside_a_file(afc: AfcService) -> None:
    await afc.set_file_contents(TEST_FILENAME, b"data")
    source = "source.txt"
    await afc.set_file_contents(source, b"data")
    impossible_path = f"{TEST_FILENAME}/dest.txt"
    try:
        with pytest.raises(AfcException):
            await afc.rename(source, impossible_path)
    finally:
        await afc.rm(TEST_FILENAME)
        await afc.rm(source)


async def test_rename_to_self(afc: AfcService) -> None:
    data = b"data"
    await afc.set_file_contents(TEST_FILENAME, data)
    await afc.rename(TEST_FILENAME, TEST_FILENAME)
    assert await afc.get_file_contents(TEST_FILENAME) == data
    await afc.rm(TEST_FILENAME)


async def test_fread_more_than_file_size(afc: AfcService) -> None:
    data = b"data"
    await afc.set_file_contents(TEST_FILENAME, data)
    h = await afc.fopen(TEST_FILENAME)
    read_data = await afc.fread(h, len(data) + 2)
    await afc.fclose(h)
    await afc.rm(TEST_FILENAME)
    assert read_data == data


async def test_fread_not_opened(afc: AfcService) -> None:
    with pytest.raises(AfcException) as e:
        await afc.fread(77, 4)
    assert e.value.status == AfcError.INVALID_ARG


async def test_fwrite_not_opened(afc: AfcService) -> None:
    with pytest.raises(AfcException) as e:
        await afc.fwrite(77, b"asdasd")
    assert e.value.status == AfcError.INVALID_ARG


async def test_file_read_write(afc: AfcService) -> None:
    body = b"data"

    await afc.set_file_contents(TEST_FILENAME, body)
    try:
        assert await afc.get_file_contents(TEST_FILENAME) == body
    finally:
        await afc.rm(TEST_FILENAME)


async def test_get_file_contents_missing_file(afc: AfcService) -> None:
    with pytest.raises(AfcFileNotFoundError):
        await afc.get_file_contents("missing_file")


async def test_dirlist(afc: AfcService):
    await afc.makedirs("test_a/test_b/test_c/test_d")
    try:
        assert [x async for x in afc.dirlist("/", 0)] == ["/"]
        dirlist = [x async for x in afc.dirlist("/", 2)]
        assert "/" in dirlist
        assert "/test_a" in dirlist
        assert "/test_a/test_b" in dirlist
        assert "/test_a/test_b/test_c" not in dirlist
        assert [x async for x in afc.dirlist("test_a", 0)] == ["test_a"]
        dirlist = [x async for x in afc.dirlist("test_a", 2)]
        assert "test_a" in dirlist
        assert "test_a/test_b" in dirlist
        assert "test_a/test_b/test_c" in dirlist
        assert "test_a/test_b/test_c/test_d" not in dirlist
    finally:
        await afc.rm("test_a")


async def test_push_pull_bigger_than_max_chunk(afc: AfcService) -> None:
    contents = b"x" * MAXIMUM_READ_SIZE * 2
    await afc.set_file_contents("test", contents)
    assert contents == await afc.get_file_contents("test")
    await afc.rm("test")


async def test_concurrent_operations(lockdown: LockdownClient) -> None:
    """
    Verify that a single AfcService instance handles multiple concurrent operations
    correctly via the packet_num demultiplexer.

    Before the background reader + packet_num demux was implemented this test failed with:
        RuntimeError: readexactly() called while another coroutine is already waiting
    """
    async with AfcService(lockdown) as afc:
        # 10 concurrent get_device_info calls — all must return non-empty dicts
        results = await asyncio.gather(*[afc.get_device_info() for _ in range(10)])
        assert all(isinstance(r, dict) and len(r) > 0 for r in results), (
            f"Expected all concurrent get_device_info calls to return non-empty dicts, got {results}"
        )

        # 5 concurrent listdir("/") — all must return the same non-empty listing
        listings = await asyncio.gather(*[afc.listdir("/") for _ in range(5)])
        assert all(isinstance(r, list) and len(r) > 0 for r in listings), (
            f"Expected all concurrent listdir calls to return non-empty lists, got {listings}"
        )
        assert all(sorted(r) == sorted(listings[0]) for r in listings), (
            f"Expected all concurrent listdir results to be identical, got {listings}"
        )



import logging
import os
import posixpath
import re
from contextlib import nullcontext
from pathlib import Path
from typing import Annotated, Optional, TextIO

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import (
    ServiceProviderDep,
    async_command,
    get_last_used_terminal_formatting,
    user_requested_colored_output,
)
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.os_trace import OsTraceService, SyslogEntry, SyslogLogLevel
from pymobiledevice3.services.syslog import SyslogService

logger = logging.getLogger(__name__)

cli = InjectingTyper(
    name="syslog",
    help="Watch syslog messages",
    no_args_is_help=True,
)


@cli.command("live-old")
@async_command
async def syslog_live_old(service_provider: ServiceProviderDep) -> None:
    """view live syslog lines in raw bytes form from old relay"""
    async with SyslogService(service_provider=service_provider) as service:
        async for line in service.watch():
            print(line)


def format_line(
    color: bool, pid: int, syslog_entry: SyslogEntry, include_label: bool, image_offset: bool = False
) -> Optional[str]:
    log_level_colors = {
        SyslogLogLevel.NOTICE.name: "white",
        SyslogLogLevel.INFO.name: "white",
        SyslogLogLevel.DEBUG.name: "green",
        SyslogLogLevel.ERROR.name: "red",
        SyslogLogLevel.FAULT.name: "red",
        SyslogLogLevel.USER_ACTION.name: "white",
    }

    syslog_pid = syslog_entry.pid
    timestamp = syslog_entry.timestamp
    level = syslog_entry.level.name
    filename = syslog_entry.filename
    image_name = posixpath.basename(syslog_entry.image_name)
    message = syslog_entry.message
    process_name = posixpath.basename(filename)
    image_offset_str = f"+0x{syslog_entry.image_offset:x}" if image_offset and image_name else ""
    label = ""

    if (pid != -1) and (syslog_pid != pid):
        return None

    if syslog_entry.label is not None:
        label = f"[{syslog_entry.label.subsystem}][{syslog_entry.label.category}]"

    if color:
        timestamp = typer.style(str(timestamp), "green")
        process_name = typer.style(process_name, "magenta")
        if len(image_name) > 0:
            image_name = typer.style(image_name, "magenta")
        if image_offset:
            image_offset_str = typer.style(image_offset_str, "blue")
        syslog_pid = typer.style(syslog_pid, "cyan")
        log_level_color = log_level_colors[level]
        level = typer.style(level, log_level_color)
        label = typer.style(label, "cyan")
        message = typer.style(message, log_level_color)

    line_format = "{timestamp} {process_name}{{{image_name}{image_offset_str}}}[{pid}] <{level}>: {message}"

    if include_label:
        line_format += f" {label}"

    line = line_format.format(
        timestamp=timestamp,
        process_name=process_name,
        image_name=image_name,
        pid=syslog_pid,
        level=level,
        message=message,
        image_offset_str=image_offset_str,
    )

    return line


def _should_skip_line(line: str, invert_match: list[str], invert_match_insensitive: list[str]) -> bool:
    for m in invert_match:
        if m in line:
            return True

    line_lower = line.lower()
    return any(m.lower() in line_lower for m in invert_match_insensitive)


def _should_keep_line(
    line: str, match: list[str], match_insensitive: list[str], match_regex: list[re.Pattern[str]]
) -> bool:
    for m in match:
        if m not in line:
            return False

    line_lower = line.lower()
    if not all(m.lower() in line_lower for m in match_insensitive):
        return False

    if not match_regex:
        return True

    return any(regex.search(line) for regex in match_regex)


def _highlight_match_filters(styled_line: str, match: list[str], match_insensitive: list[str]) -> str:
    for m in match:
        styled_line = styled_line.replace(m, typer.style(m, bold=True, underline=True))

    for m in match_insensitive:
        m = m.lower()
        start = styled_line.lower().index(m)
        end = start + len(m)
        last_color_formatting = get_last_used_terminal_formatting(styled_line[:start])
        styled_line = (
            styled_line[:start]
            + typer.style(styled_line[start:end], bold=True, underline=True)
            + last_color_formatting
            + styled_line[end:]
        )

    return styled_line


def _highlight_regex_filters(styled_line: str, match_regex: list[re.Pattern[str]]) -> str:
    def replace(m):
        if len(m.groups()):
            return styled_line.replace(m.group(1), typer.style(m.group(1), bold=True, underline=True))
        return ""

    for regex in match_regex:
        if regex.search(styled_line):
            styled_line = re.sub(regex, replace, styled_line)

    return styled_line


async def syslog_live(
    service_provider: LockdownServiceProvider,
    out: Optional[TextIO],
    pid: int,
    process_name: Optional[str],
    match: list[str],
    invert_match: list[str],
    match_insensitive: list[str],
    invert_match_insensitive: list[str],
    include_label: bool,
    regex: list[str],
    insensitive_regex: list[str],
    image_offset: bool = False,
    start_after: Optional[str] = None,
) -> None:
    match_regex = [re.compile(f".*({r}).*", re.DOTALL) for r in regex]
    match_regex += [re.compile(f".*({r}).*", re.IGNORECASE | re.DOTALL) for r in insensitive_regex]
    started = start_after is None

    if start_after is not None:
        print(f'Waiting for "{start_after}" ...', flush=True)

    should_color_output = user_requested_colored_output()

    async for syslog_entry in OsTraceService(lockdown=service_provider).syslog(pid=pid):
        if process_name and posixpath.basename(syslog_entry.filename) != process_name:
            continue

        line = format_line(False, pid, syslog_entry, include_label, image_offset)
        styled_line = format_line(should_color_output, pid, syslog_entry, include_label, image_offset)

        if line is None or styled_line is None:
            continue

        if not started:
            if start_after not in line:
                continue
            started = True

        if _should_skip_line(line, invert_match, invert_match_insensitive):
            continue

        if not _should_keep_line(line, match, match_insensitive, match_regex):
            continue

        if should_color_output:
            styled_line = _highlight_match_filters(styled_line, match, match_insensitive)
            styled_line = _highlight_regex_filters(styled_line, match_regex)

        print(styled_line, flush=True)

        if out:
            if should_color_output:
                print(styled_line, file=out, flush=True)
            else:
                print(line, file=out, flush=True)


@cli.command("live")
@async_command
async def cli_syslog_live(
    service_provider: ServiceProviderDep,
    out: Annotated[
        Optional[Path],
        typer.Option(
            "--out",
            "-o",
            help="log file",
        ),
    ] = None,
    pid: Annotated[
        int,
        typer.Option(help="pid to filter. -1 for all"),
    ] = -1,
    process_name: Annotated[
        Optional[str],
        typer.Option(
            "--process-name",
            "-pn",
            help="process name to filter",
        ),
    ] = None,
    match: Annotated[
        Optional[list[str]],
        typer.Option(
            "--match",
            "-m",
            help="filter only logs matching this expression (repeatable; all must match - conjunction)",
        ),
    ] = None,
    invert_match: Annotated[
        Optional[list[str]],
        typer.Option(
            "--invert-match",
            "-v",
            help="filter only logs not matching this expression (repeatable; any match excludes - disjunction)",
        ),
    ] = None,
    match_insensitive: Annotated[
        Optional[list[str]],
        typer.Option(
            "--match-insensitive",
            "-mi",
            help="filter only logs matching this expression, case-insensitively "
            "(repeatable; all must match - conjunction)",
        ),
    ] = None,
    invert_match_insensitive: Annotated[
        Optional[list[str]],
        typer.Option(
            "--invert-match-insensitive",
            "-vi",
            help="filter only logs not matching this expression, case-insensitively "
            "(repeatable; any match excludes - disjunction)",
        ),
    ] = None,
    include_label: Annotated[
        bool,
        typer.Option(
            "--label",
            help="should include label",
        ),
    ] = False,
    regex: Annotated[
        Optional[list[str]],
        typer.Option(
            "--regex",
            "-e",
            help="filter only lines matching given regex (repeatable; any match includes - disjunction)",
        ),
    ] = None,
    insensitive_regex: Annotated[
        Optional[list[str]],
        typer.Option(
            "--insensitive-regex",
            "-ei",
            help="filter only lines matching given regex, case-insensitively "
            "(repeatable; any match includes - disjunction)",
        ),
    ] = None,
    image_offset: Annotated[
        bool,
        typer.Option(
            "--image-offset",
            "-io",
            help="Include image offset in log line",
        ),
    ] = False,
    start_after: Annotated[
        Optional[str],
        typer.Option(
            "--start-after",
            help="Start printing only after this string is seen",
        ),
    ] = None,
) -> None:
    """view live syslog lines"""

    with out.open("wt") if out else nullcontext() as out_file:
        await syslog_live(
            service_provider,
            out_file,
            pid,
            process_name,
            match or [],
            invert_match or [],
            match_insensitive or [],
            invert_match_insensitive or [],
            include_label,
            regex or [],
            insensitive_regex or [],
            image_offset,
            start_after,
        )


@cli.command("collect")
@async_command
async def syslog_collect(
    service_provider: ServiceProviderDep,
    out: Annotated[
        Path,
        typer.Argument(
            exists=False,
            dir_okay=True,
            file_okay=False,
        ),
    ],
    size_limit: Annotated[
        Optional[int],
        typer.Option(
            "--size-limit",
            help="Maximum size in bytes of logarchive",
        ),
    ] = None,
    age_limit: Annotated[
        Optional[int],
        typer.Option(
            "--age-limit",
            help="Maximum age in days",
        ),
    ] = None,
    start_time: Annotated[
        Optional[int],
        typer.Option(
            "--start-time",
            help="Start time of logarchive as a unix timestamp",
        ),
    ] = None,
) -> None:
    """
    Collect the system logs into a .logarchive that can be viewed later with tools such as log or Console.
    If the filename doesn't exist, system_logs.logarchive will be created in the given directory.
    """
    if not os.path.exists(out):
        os.makedirs(out)

    if out.suffix != ".logarchive":
        logger.warning(
            "given out path doesn't end with a .logarchive - consider renaming to be able to view "
            "the file with the likes of the Console.app and the `log show` utilities"
        )

    await OsTraceService(lockdown=service_provider).collect(
        str(out), size_limit=size_limit, age_limit=age_limit, start_time=start_time
    )



import logging

from pymobiledevice3.dtx import DTXService, dtx_method
from pymobiledevice3.dtx_service import DtxService
from pymobiledevice3.exceptions import PyMobileDevice3Exception


class ConditionInducerService(DTXService):
    IDENTIFIER = "com.apple.instruments.server.services.ConditionInducer"

    @dtx_method("availableConditionInducers")
    async def available_condition_inducers(self) -> list: ...

    @dtx_method("enableConditionWithIdentifier:profileIdentifier:")
    async def enable_condition_with_identifier_profile_identifier_(
        self, condition_identifier: str, profile_identifier: str
    ) -> None: ...

    @dtx_method("disableActiveCondition", expects_reply=False)
    async def disable_active_condition(self) -> None: ...


class ConditionInducer(DtxService[ConditionInducerService]):
    def __init__(self, dvt):
        super().__init__(dvt)
        self.logger = logging.getLogger(__name__)

    async def list(self) -> list[dict]:
        return await self.service.available_condition_inducers()

    async def set(self, profile_identifier):
        for group in await self.list():
            for profile in group.get("profiles"):
                if profile_identifier == profile.get("identifier"):
                    self.logger.info(profile.get("description"))
                    await self.service.enable_condition_with_identifier_profile_identifier_(
                        group.get("identifier"), profile.get("identifier")
                    )
                    return
        raise PyMobileDevice3Exception("Invalid profile identifier")

    async def clear(self):
        await self.service.disable_active_condition()



import posixpath
from collections.abc import AsyncGenerator
from contextlib import suppress
from datetime import datetime, timezone
from json import JSONDecodeError
from pathlib import Path
from typing import Optional

import pytest
import pytest_asyncio

from pymobiledevice3.exceptions import AfcFileNotFoundError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.crash_reports import CrashReportsManager

BASENAME = "__pymobiledevice3_tests"
PATH_COMPONENT = f"/{BASENAME}"


@pytest_asyncio.fixture(scope="function")
async def crash_manager(lockdown: LockdownClient):
    async with CrashReportsManager(lockdown) as crash_manager:
        yield crash_manager


@pytest_asyncio.fixture(scope="function")
async def remote_temp_directory(lockdown: LockdownClient) -> AsyncGenerator[str]:
    async with CrashReportsManager(lockdown) as crash_manager:
        with suppress(AfcFileNotFoundError):
            await crash_manager.afc.rm(BASENAME)
        await crash_manager.afc.makedirs(PATH_COMPONENT)
    yield PATH_COMPONENT
    async with CrashReportsManager(lockdown) as crash_manager:
        with suppress(AfcFileNotFoundError):
            await crash_manager.afc.rm(BASENAME)


async def _create_crash_report(crash_manager: CrashReportsManager, filename: str, name: Optional[str] = None) -> None:
    creation_time = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f +0000")
    name_str = posixpath.splitext(posixpath.basename(filename))[0] if name is None else name

    await crash_manager.afc.set_file_contents(
        filename,
        (
            '{"bug_type":"999","incident_id":"'
            + name_str
            + '","timestamp":"'
            + creation_time
            + '","name":"'
            + name_str
            + '"}\n{"payload":"ok"}'
        ).encode(),
    )


def _nested_paths(root: str, depth: int) -> list[str]:
    paths = [root]
    for _ in range(1, depth):
        paths.append(posixpath.join(paths[-1], BASENAME))
    return paths


async def test_ls_default(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    assert remote_temp_directory in await crash_manager.ls()


async def test_ls_path(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    remote_nested_directory = posixpath.join(remote_temp_directory, BASENAME)
    await crash_manager.afc.makedirs(remote_nested_directory)
    assert remote_nested_directory in await crash_manager.ls(path=remote_temp_directory)


@pytest.mark.parametrize("depth", [2, 3, 4])
async def test_ls_depth(crash_manager: CrashReportsManager, remote_temp_directory: str, depth: int) -> None:
    remote_path_list = _nested_paths(remote_temp_directory, depth)
    await crash_manager.afc.makedirs(remote_path_list[-1])
    crash_list = await crash_manager.ls(depth=depth)
    for item in remote_path_list:
        assert item in crash_list


async def test_ls_depth_minus_one(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    remote_path_list = _nested_paths(remote_temp_directory, 3)
    await crash_manager.afc.makedirs(remote_path_list[-1])
    crash_list = await crash_manager.ls(depth=-1)
    for path in remote_path_list:
        assert path in crash_list


async def test_clear(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "clear_test_report.ips"))
    await crash_manager.afc.makedirs(posixpath.join(remote_temp_directory, "nested"))
    await _create_crash_report(
        crash_manager, posixpath.join(remote_temp_directory, "nested", "clear_test_nested_report.ips")
    )

    files = await crash_manager.afc.listdir(remote_temp_directory)
    assert len(files) > 0

    await crash_manager.clear(remote_temp_directory)

    files = await crash_manager.afc.listdir(remote_temp_directory)
    assert len(files) == 0


async def test_pull_file(crash_manager: CrashReportsManager, remote_temp_directory: str, tmp_path: Path) -> None:
    remote_file = posixpath.join(remote_temp_directory, "pull_test_report.ips")
    await _create_crash_report(crash_manager, remote_file)

    local_output_dir = tmp_path
    await crash_manager.pull(str(local_output_dir), remote_file)

    local_file = local_output_dir / "pull_test_report.ips"
    assert local_file.exists()


async def test_pull_directory(crash_manager: CrashReportsManager, remote_temp_directory: str, tmp_path: Path) -> None:
    remote_file = posixpath.join(remote_temp_directory, "pull_test_report.ips")
    remote_nested_directory = posixpath.join(remote_temp_directory, "nested")
    remote_nested_file = posixpath.join(remote_nested_directory, "pull_test_nested_report.ips")
    await _create_crash_report(crash_manager, remote_file)
    await crash_manager.afc.makedirs(remote_nested_directory)
    await _create_crash_report(crash_manager, remote_nested_file)

    local_output_dir = tmp_path
    await crash_manager.pull(str(local_output_dir), remote_temp_directory)

    pulled_directories = [path for path in local_output_dir.iterdir() if path.is_dir()]
    assert len(pulled_directories) == 1
    local_pulled_directory = pulled_directories[0]
    assert local_pulled_directory.is_dir()
    assert (local_pulled_directory / "nested").is_dir()
    assert (local_pulled_directory / "nested" / "pull_test_nested_report.ips").is_file()
    assert (local_pulled_directory / "pull_test_report.ips").is_file()


async def test_parse(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    filename = posixpath.join(remote_temp_directory, "parse_test_report.ips")
    await _create_crash_report(crash_manager, filename)

    parsed = await crash_manager.parse(filename)
    assert parsed.filename == filename
    assert parsed.name == "parse_test_report"
    assert parsed.incident_id == "parse_test_report"


async def test_parse_invalid_raises(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    filename = posixpath.join(remote_temp_directory, "invalid_report.ips")
    await crash_manager.afc.set_file_contents(filename, b"not-json\n{}")

    with pytest.raises(JSONDecodeError):
        await crash_manager.parse(filename)


@pytest.mark.parametrize(
    ("end_time", "return_value"),
    ((-1, True), (0, True), (float("inf"), False), (None, False)),
)
def test_check_timeout(crash_manager: CrashReportsManager, end_time: int, return_value: bool) -> None:
    assert crash_manager._check_timeout(end_time) is return_value


async def test_parse_latest(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    older = posixpath.join(remote_temp_directory, "quite_old.ips")
    newer = posixpath.join(remote_temp_directory, "quite_new.ips")

    await _create_crash_report(crash_manager, older)
    await _create_crash_report(crash_manager, newer)

    parsed = await crash_manager.parse_latest(path=remote_temp_directory)
    assert len(parsed) == 1
    assert parsed[0].filename == newer


@pytest.mark.parametrize(
    ("match", "match_insensitive"),
    (
        ([r"ite_old"], None),
        ([r"old", r"quite"], None),
        (None, [r"iTe_Old"]),
        (None, [r"olD", r"QUIte"]),
    ),
)
async def test_parse_latest_filter_variants(
    crash_manager: CrashReportsManager,
    remote_temp_directory: str,
    match: Optional[list[str]],
    match_insensitive: Optional[list[str]],
) -> None:
    oldest = posixpath.join(remote_temp_directory, "quite_old.ips")
    await _create_crash_report(crash_manager, oldest)
    for i in range(10):
        await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, f"too_old_{i}.ips"))
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "quite_new.ips"))

    parsed = await crash_manager.parse_latest(
        path=remote_temp_directory, match=match, match_insensitive=match_insensitive
    )

    assert len(parsed) == 1
    assert parsed[0].filename == oldest


async def test_parse_latest_match_and_match_insensitive_combined(
    crash_manager: CrashReportsManager, remote_temp_directory: str
) -> None:
    report_paths = [posixpath.join(remote_temp_directory, f"one_two{i}.ips") for i in range(5)]
    report_paths += [posixpath.join(remote_temp_directory, f"Two_Three{i}.ips") for i in range(5)]

    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "just_here.ips"))
    for report_path in report_paths:
        await _create_crash_report(crash_manager, report_path)
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "three_sensitive.ips"))
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "Three_but_no_t_w_o.ips"))

    parsed = await crash_manager.parse_latest(path=remote_temp_directory, match=["Three"], match_insensitive=["two"])
    assert len(parsed) == 1
    assert parsed[0].filename == report_paths[-1]


async def test_parse_latest_ignores_nested_matches(
    crash_manager: CrashReportsManager, remote_temp_directory: str
) -> None:
    matching_dir = posixpath.join(remote_temp_directory, "report_dir")
    another_matching_dir = posixpath.join(remote_temp_directory, "report")
    matching_file = posixpath.join(matching_dir, "report_file.ips")

    await crash_manager.afc.makedirs(matching_dir)
    await crash_manager.afc.makedirs(another_matching_dir)
    await _create_crash_report(crash_manager, matching_file)

    with pytest.raises(ValueError, match="No reports found"):
        await crash_manager.parse_latest(path=remote_temp_directory, match=[r"report"], count=3)


async def test_parse_latest_prefers_top_level_matches(
    crash_manager: CrashReportsManager, remote_temp_directory: str
) -> None:
    matching_dir = posixpath.join(remote_temp_directory, "report_dir")
    another_matching_dir = posixpath.join(remote_temp_directory, "report")
    matching_file = posixpath.join(matching_dir, "report_file.ips")
    top_level_file = posixpath.join(remote_temp_directory, "report_top_level.ips")

    await crash_manager.afc.makedirs(matching_dir)
    await crash_manager.afc.makedirs(another_matching_dir)
    await _create_crash_report(crash_manager, matching_file)
    await _create_crash_report(crash_manager, top_level_file)

    parsed = await crash_manager.parse_latest(path=remote_temp_directory, match=[r"report"], count=3)
    assert len(parsed) == 1
    assert parsed[0].filename == top_level_file


async def test_parse_latest_count(crash_manager: CrashReportsManager, remote_temp_directory: str) -> None:
    count = 3
    report_paths = [posixpath.join(remote_temp_directory, f"report{i}.ips") for i in range(2 * count)]
    for report_path in report_paths:
        await _create_crash_report(crash_manager, report_path)

    parsed = await crash_manager.parse_latest(path=remote_temp_directory, match=[r"report"], count=count)
    assert [p.filename for p in parsed] == list(reversed(report_paths))[:count]


async def test_parse_latest_count_larger_than_available_reports(
    crash_manager: CrashReportsManager, remote_temp_directory: str
) -> None:
    report_paths = [posixpath.join(remote_temp_directory, f"report{i}.ips") for i in range(5)]
    more_report_paths = [posixpath.join(remote_temp_directory, f"noiseee{i}.ips") for i in range(10)]
    for report_path in report_paths + more_report_paths:
        await _create_crash_report(crash_manager, report_path)

    parsed = await crash_manager.parse_latest(path=remote_temp_directory, match=[r"report"], count=100)
    assert [p.filename for p in parsed] == list(reversed(report_paths))


async def test_parse_latest_count_match_and_match_insensitive_combined(
    crash_manager: CrashReportsManager, remote_temp_directory: str
) -> None:
    report_paths = [posixpath.join(remote_temp_directory, f"One_Two{i}.ips") for i in range(5)]
    more_report_paths = [posixpath.join(remote_temp_directory, f"one_two{i}.ips") for i in range(5)]

    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "just_here.ips"))
    for report_path in report_paths + more_report_paths:
        await _create_crash_report(crash_manager, report_path)
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "one_but_not_T-w-o.ips"))
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "Two_but_not_o-n-e.ips"))

    parsed = await crash_manager.parse_latest(
        path=remote_temp_directory, match=["Two"], match_insensitive=["one"], count=len(report_paths)
    )
    assert [p.filename for p in parsed] == list(reversed(report_paths))


async def test_parse_latest_invalid_count_raises(crash_manager: CrashReportsManager) -> None:
    with pytest.raises(ValueError):
        await crash_manager.parse_latest(count=0)


@pytest.mark.parametrize(
    ("match", "match_insensitive"),
    (
        (["missing"], None),
        (None, ["also-missing"]),
        (["missing"], ["also-missing"]),
    ),
)
async def test_parse_latest_no_matches(
    crash_manager: CrashReportsManager,
    remote_temp_directory: str,
    match: Optional[list[str]],
    match_insensitive: Optional[list[str]],
) -> None:
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "hello.ips"))
    await _create_crash_report(crash_manager, posixpath.join(remote_temp_directory, "world.ips"))

    with pytest.raises(ValueError):
        await crash_manager.parse_latest(path=remote_temp_directory, match=match, match_insensitive=match_insensitive)



# Async, dependency-light mDNS browser returning dataclasses with per-address interface names.
# Works for any DNS-SD type, e.g. "_remoted._tcp.local."
# - Uses ifaddr (optional) to map IPs -> local interfaces; otherwise iface will be None.

import asyncio
import contextlib
import ipaddress
import socket
import struct
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional

import ifaddr  # pip install ifaddr

from pymobiledevice3.osu.os_utils import get_os_utils

REMOTEPAIRING_SERVICE_NAME = "_remotepairing._tcp.local."
REMOTEPAIRING_MANUAL_PAIRING_SERVICE_NAME = "_remotepairing-manual-pairing._tcp.local."
MOBDEV2_SERVICE_NAME = "_apple-mobdev2._tcp.local."
REMOTED_SERVICE_NAME = "_remoted._tcp.local."
OSUTILS = get_os_utils()
DEFAULT_BONJOUR_TIMEOUT = OSUTILS.bonjour_timeout

MDNS_PORT = 5353
MDNS_MCAST_V4 = "224.0.0.251"
MDNS_MCAST_V6 = "ff02::fb"

QTYPE_A = 1
QTYPE_PTR = 12
QTYPE_TXT = 16
QTYPE_AAAA = 28
QTYPE_SRV = 33

CLASS_IN = 0x0001
CLASS_QU = 0x8000  # unicast-response bit (we use multicast queries)


# ---------------- Dataclasses ----------------


# --- Dataclass decorator shim (adds slots only on 3.10+)
def dataclass_compat(*d_args, **d_kwargs):
    if sys.version_info < (3, 10):
        d_kwargs.pop("slots", None)  # ignore on 3.9
    return dataclass(*d_args, **d_kwargs)


@dataclass_compat(slots=True)
class Address:
    ip: str
    iface: str  # local interface name (e.g., "en0"), or None if unknown

    @property
    def full_ip(self) -> str:
        if self.iface and self.ip.lower().startswith("fe80:"):
            return f"{self.ip}%{self.iface}"
        return self.ip


@dataclass_compat(slots=True)
class ServiceInstance:
    instance: str  # "<Instance Name>._type._proto.local."
    host: Optional[str]  # "host.local" (without trailing dot), or None if unresolved
    port: Optional[int]  # SRV port
    addresses: list[Address] = field(default_factory=list)  # IPs with interface names
    properties: dict[str, str] = field(default_factory=dict)  # TXT key/values


# ---------------- DNS helpers ----------------


def encode_name(name: str) -> bytes:
    name = name.rstrip(".")
    out = bytearray()
    for label in name.split(".") if name else []:
        b = label.encode("utf-8")
        if len(b) > 63:
            raise ValueError("label too long")
        out.append(len(b))
        out += b
    out.append(0)
    return bytes(out)


def decode_name(data: bytes, off: int) -> tuple[str, int]:
    labels = []
    jumped = False
    orig_end = off
    for _ in range(128):  # loop guard
        if off >= len(data):
            break
        length = data[off]
        if length == 0:
            off += 1
            break
        if (length & 0xC0) == 0xC0:
            if off + 1 >= len(data):
                raise ValueError("truncated name pointer")
            ptr = ((length & 0x3F) << 8) | data[off + 1]
            if ptr >= len(data):
                raise ValueError("bad name pointer")
            if not jumped:
                orig_end = off + 2
            off = ptr
            jumped = True
            continue
        off += 1
        end = off + length
        if end > len(data):
            raise ValueError("truncated label")
        labels.append(data[off:end].decode("utf-8", errors="replace"))
        off = end
    return ".".join(labels) + ".", (orig_end if jumped else off)


def build_query(name: str, qtype: int, unicast: bool = False) -> bytes:
    hdr = struct.pack("!HHHHHH", 0, 0, 1, 0, 0, 0)  # TXID=0, flags=0, 1 question
    qclass = CLASS_IN | (CLASS_QU if unicast else 0)
    return hdr + encode_name(name) + struct.pack("!HH", qtype, qclass)


def parse_rr(data: bytes, off: int):
    name, off = decode_name(data, off)
    if off + 10 > len(data):
        raise ValueError("truncated RR header")
    rtype, rclass, ttl, rdlen = struct.unpack("!HHIH", data[off : off + 10])
    off += 10
    rdata = data[off : off + rdlen]
    off += rdlen

    rr = {"name": name, "type": rtype, "class": rclass & 0x7FFF, "ttl": ttl}
    if rtype == QTYPE_PTR:
        target, _ = decode_name(data, off - rdlen)
        rr["ptrdname"] = target
    elif rtype == QTYPE_SRV and rdlen >= 6:
        priority, weight, port = struct.unpack("!HHH", rdata[:6])
        target, _ = decode_name(data, off - rdlen + 6)
        rr.update({"priority": priority, "weight": weight, "port": port, "target": target})
    elif rtype == QTYPE_TXT:
        kv = {}
        i = 0
        while i < rdlen:
            b = rdata[i]
            i += 1
            seg = rdata[i : i + b]
            i += b
            if not seg:
                continue
            if b"=" in seg:
                k, v = seg.split(b"=", 1)
                kv[k.decode()] = v.decode(errors="replace")
            else:
                kv[seg.decode()] = ""
        rr["txt"] = kv
    elif rtype == QTYPE_A and rdlen == 4:
        rr["address"] = socket.inet_ntop(socket.AF_INET, rdata)
    elif rtype == QTYPE_AAAA and rdlen == 16:
        rr["address"] = socket.inet_ntop(socket.AF_INET6, rdata)
    else:
        rr["raw"] = rdata
    return rr, off


def parse_mdns_message(data: bytes):
    if len(data) < 12:
        return []
    _, _, qd, an, ns, ar = struct.unpack("!HHHHHH", data[:12])
    off = 12
    for _ in range(qd):
        _, off = decode_name(data, off)
        off += 4
    rrs = []
    for _ in range(an + ns + ar):
        rr, off = parse_rr(data, off)
        rrs.append(rr)
    return rrs


# ---------------- Interface mapping helpers ----------------


class _Adapters:
    def __init__(self):
        self.adapters = ifaddr.get_adapters() if ifaddr is not None else []

    def pick_iface_for_ip(self, ip_str: str, family: int, v6_scopeid: Optional[int]) -> Optional[str]:
        # Prefer scope id for IPv6 link-local
        if family == socket.AF_INET6 and ip_str.lower().startswith("fe80:") and v6_scopeid:
            try:
                return socket.if_indextoname(v6_scopeid)
            except OSError:
                pass

        # Otherwise, try to match destination ip to local subnet via ifaddr
        if not self.adapters:
            return None
        ip = ipaddress.ip_address(ip_str)
        best = (None, -1)  # (name, prefix_len)
        for ad in self.adapters:
            for ipn in ad.ips:
                if isinstance(ipn.ip, str):
                    # IPv4
                    fam = socket.AF_INET
                    ipn_ip = ipn.ip
                else:
                    # IPv6
                    fam = socket.AF_INET6
                    ipn_ip = ipn.ip[0]
                if fam != family:
                    continue
                net = ipaddress.ip_network(f"{ipn_ip}/{ipn.network_prefix}", strict=False)
                if ip in net and ipn.network_prefix > best[1]:
                    best = (ad.nice_name or ad.name, ipn.network_prefix)
        return best[0]


# ---------------- async sockets ----------------


class _DatagramProtocol(asyncio.DatagramProtocol):
    def __init__(self, queue: asyncio.Queue):
        self.queue = queue

    def datagram_received(self, data, addr):
        # addr: IPv4 -> (host, port); IPv6 -> (host, port, flowinfo, scopeid)
        self.queue.put_nowait((data, addr))


async def _bind_ipv4(queue: asyncio.Queue):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if hasattr(socket, "SO_REUSEPORT"):
        with contextlib.suppress(OSError):
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    s.bind(("0.0.0.0", MDNS_PORT))
    try:
        mreq = struct.pack("=4s4s", socket.inet_aton(MDNS_MCAST_V4), socket.inet_aton("0.0.0.0"))
        s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    except OSError:
        pass
    transport, _ = await asyncio.get_running_loop().create_datagram_endpoint(lambda: _DatagramProtocol(queue), sock=s)
    return transport, s


async def _bind_ipv6_all_ifaces(queue: asyncio.Queue):
    s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if hasattr(socket, "SO_REUSEPORT"):
        with contextlib.suppress(OSError):
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    s.bind(("::", MDNS_PORT))
    grp = socket.inet_pton(socket.AF_INET6, MDNS_MCAST_V6)
    for ifindex, _ in socket.if_nameindex():
        mreq6 = grp + struct.pack("@I", ifindex)
        try:
            s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, mreq6)
        except OSError:
            continue
    transport, _ = await asyncio.get_running_loop().create_datagram_endpoint(lambda: _DatagramProtocol(queue), sock=s)
    return transport, s


async def _open_mdns_sockets():
    queue = asyncio.Queue()
    transports: list[tuple[asyncio.BaseTransport, socket.socket]] = []
    t4, s4 = await _bind_ipv4(queue)
    transports.append((t4, s4))
    t6, s6 = await _bind_ipv6_all_ifaces(queue)
    transports.append((t6, s6))
    if not transports:
        raise RuntimeError("Failed to open mDNS sockets (UDP/5353)")
    return transports, queue


async def _send_query_all(transports, pkt: bytes):
    for transport, sock in transports:
        if sock.family == socket.AF_INET:
            transport.sendto(pkt, (MDNS_MCAST_V4, MDNS_PORT))
        else:
            # Send once per iface index for better reachability
            for ifindex, _ in socket.if_nameindex():
                transport.sendto(pkt, (MDNS_MCAST_V6, MDNS_PORT, 0, ifindex))


# ---------------- Public API ----------------


async def browse_service(service_type: str, timeout: float = 4.0) -> list[ServiceInstance]:
    """
    Discover a DNS-SD/mDNS service type (e.g. "_remoted._tcp.local.") on the local network.

    Returns: List[ServiceInstance] with Address(ip, iface) entries.
    """
    if not service_type.endswith("."):
        service_type += "."

    transports, queue = await _open_mdns_sockets()
    adapters = _Adapters()

    ptr_targets: set[str] = set()
    srv_map: dict[str, list[dict]] = defaultdict(list)  # instance_name -> list of {"target", "port"}
    txt_map: dict[str, dict] = {}
    host_addrs: dict[str, list[Address]] = defaultdict(list)  # host -> list[(ip, iface)]

    def _record_addr(rr_name: str, ip_str: str, pkt_addr):
        # Determine family and possible scopeid from the packet that delivered this RR
        family = socket.AF_INET6 if ":" in ip_str else socket.AF_INET
        scopeid = None
        if isinstance(pkt_addr, tuple) and len(pkt_addr) == 4:  # IPv6 remote tuple
            scopeid = pkt_addr[3]
        iface = adapters.pick_iface_for_ip(ip_str, family, scopeid)
        if iface is None:
            return
        # Avoid duplicates for the same host/ip
        existing = host_addrs[rr_name]
        if not any(a.ip == ip_str for a in existing):
            existing.append(Address(ip=ip_str, iface=iface))

    try:
        await _send_query_all(transports, build_query(service_type, QTYPE_PTR, unicast=False))
        loop = asyncio.get_running_loop()
        end = loop.time() + timeout
        while loop.time() < end:
            try:
                data, pkt_addr = await asyncio.wait_for(queue.get(), timeout=end - loop.time())
            except asyncio.TimeoutError:
                break
            for rr in parse_mdns_message(data):
                t = rr.get("type")
                if t == QTYPE_PTR and rr.get("name") == service_type:
                    ptr_targets.add(rr.get("ptrdname"))
                elif t == QTYPE_SRV:
                    srv_map[rr["name"]].append({"target": rr.get("target"), "port": rr.get("port")})
                elif t == QTYPE_TXT:
                    # TODO: This could possibly mix the properties of multiple TXT records for the same instance.
                    #       However, it's currently unused.
                    txt_map[rr["name"]] = rr.get("txt", {})
                elif (t == QTYPE_A and rr.get("address")) or (t == QTYPE_AAAA and rr.get("address")):
                    _record_addr(rr["name"], rr["address"], pkt_addr)
    finally:
        for transport, _ in transports:
            transport.close()

    # Assemble dataclasses
    results: list[ServiceInstance] = []
    for inst in sorted(ptr_targets):
        srv_entries = srv_map.get(inst, [])
        props = txt_map.get(inst, {})
        for srv in srv_entries:
            target = srv.get("target")
            host = (target[:-1] if target and target.endswith(".") else target) or None
            addrs = host_addrs.get(target, []) if target else []
            results.append(
                ServiceInstance(
                    instance=inst,
                    host=host,
                    port=srv.get("port"),
                    addresses=addrs,
                    properties=props,
                )
            )
    return results


async def browse_remoted(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> list[ServiceInstance]:
    return await browse_service(REMOTED_SERVICE_NAME, timeout=timeout)


async def browse_mobdev2(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> list[ServiceInstance]:
    return await browse_service(MOBDEV2_SERVICE_NAME, timeout=timeout)


async def browse_remotepairing(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> list[ServiceInstance]:
    return await browse_service(REMOTEPAIRING_SERVICE_NAME, timeout=timeout)


async def browse_remotepairing_manual_pairing(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> list[ServiceInstance]:
    return await browse_service(REMOTEPAIRING_MANUAL_PAIRING_SERVICE_NAME, timeout=timeout)



import asyncio
from typing import Optional

import pytest

from pymobiledevice3.exceptions import DvtDirListError, UnrecognizedSelectorError
from pymobiledevice3.services.dvt.instruments.application_listing import ApplicationListing
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl


async def get_process_data(device_info: DeviceInfo, name: str):
    processes = await device_info.proclist()
    return next(process for process in processes if process["name"] == name)


@pytest.mark.asyncio
async def test_ls(dvt) -> None:
    """
    Test listing a directory.
    """
    async with DeviceInfo(dvt) as device_info:
        ls = set(await device_info.ls("/"))
    assert {"usr", "bin", "etc", "var", "private", "Applications", "Developer"} <= ls


@pytest.mark.asyncio
async def test_ls_failure(dvt) -> None:
    """
    Test listing a directory.
    """
    async with DeviceInfo(dvt) as device_info:
        with pytest.raises(DvtDirListError):
            await device_info.ls("Directory that does not exist")


@pytest.mark.asyncio
async def test_proclist(dvt) -> None:
    """
    Test listing processes.
    """
    async with DeviceInfo(dvt) as device_info:
        lockdownd = await get_process_data(device_info, "lockdownd")
    assert lockdownd["realAppName"] == "/usr/libexec/lockdownd"
    assert not lockdownd["isApplication"]


@pytest.mark.asyncio
async def test_applist(dvt) -> None:
    """
    Test listing applications.
    """
    async with ApplicationListing(dvt) as application_listing:
        apps = await application_listing.applist()
    safari = next(app for app in apps if app["DisplayName"] == "StocksWidget")
    assert safari["CFBundleIdentifier"] == "com.apple.stocks.widget"
    assert safari["Restricted"] == 1
    assert safari["Type"] == "PluginKit"


@pytest.mark.asyncio
async def test_memlimitoff(dvt) -> None:
    """
    Test disabling memory limit.
    """
    async with DeviceInfo(dvt) as device_info:
        process = await get_process_data(device_info, "SpringBoard")
    async with ProcessControl(dvt) as process_control:
        await process_control.disable_memory_limit_for_pid(process["pid"])


@pytest.mark.asyncio
async def test_kill(dvt, service_provider) -> None:
    """
    Test killing a process.
    """
    async with DeviceInfo(dvt) as device_info:
        aggregated = await get_process_data(device_info, "SpringBoard")
    async with ProcessControl(dvt) as process_control:
        await process_control.kill(aggregated["pid"])
    # give the os some time to start the process again
    await asyncio.sleep(3)
    async with type(dvt)(service_provider) as second_dvt, DeviceInfo(second_dvt) as device_info:
        aggregated_after_kill = await get_process_data(device_info, "SpringBoard")
    if "startDate" in aggregated:
        assert aggregated["startDate"] < aggregated_after_kill["startDate"]


@pytest.mark.asyncio
async def test_launch(dvt, service_provider) -> None:
    """
    Test launching a process.
    """
    async with ProcessControl(dvt) as process_control:
        pid = await process_control.launch("com.apple.mobilesafari")
    assert pid
    async with type(dvt)(service_provider) as second_dvt, DeviceInfo(second_dvt) as device_info:
        processes = await device_info.proclist()
    for process in processes:
        if pid == process["pid"]:
            assert process["name"] == "MobileSafari"


@pytest.mark.asyncio
@pytest.mark.skip("this test is brittle, yet it was a quick way to test callbacks")
async def test_observe(dvt, service_provider) -> None:
    """
    Test observing a launched process and receiving its termination callback.
    """
    process_terminated: asyncio.Future[tuple[int, int, Optional[int]]] = asyncio.get_event_loop().create_future()
    signature_received: asyncio.Future[dict] = asyncio.get_event_loop().create_future()

    async def on_process_terminated(pid: int, exit_code: int, crashing_signal: Optional[int]) -> None:
        process_terminated.set_result((pid, exit_code, crashing_signal))

    async def on_tracking_signature(obj: dict) -> None:
        if not signature_received.done():
            signature_received.set_result(obj)

    async with ProcessControl(dvt) as process_control, DeviceInfo(dvt) as device_info:
        process_control.service._on_process_terminated = on_process_terminated

        # manually add a @dtx_on_invoke
        device_info.service._on_tracking_signature = on_tracking_signature
        device_info.service._dtx_dispatch["trackProcess:"] = "_on_tracking_signature"
        device_info.service._channel.on_invoke = device_info.service.__on_dispatch__

        pid = await process_control.launch("com.apple.mobilesafari")
        assert pid
        await process_control.service.startObservingPid_(pid)
        x = await device_info.service.symbolicator_signature_for_pid_tracking_selector_(pid, "trackProcess:")
        assert x is None or isinstance(x, bytes), (
            f"Expected symbolicator signature to be bytes or None, got {type(x).__name__}"
        )
        obj = await asyncio.wait_for(signature_received, 3)
        await process_control.kill(pid)
        assert obj is not None, "Expected a tracking object but got None"
        assert type(obj) is dict, f"Expected a tracking object of type dict but got {type(obj).__name__}"
        assert "pid" in obj, "Expected 'pid' key in tracking object"
        assert "signature" in obj, "Expected 'signature' key in tracking object"
        assert isinstance(obj["signature"], bytes), (
            f"Expected 'signature' value to be bytes but got {type(obj['signature']).__name__}"
        )
        terminated_pid, _exit_code, _crashing_signal = await asyncio.wait_for(process_terminated, 1)
        assert terminated_pid == pid


@pytest.mark.asyncio
async def test_system_information(dvt) -> None:
    """
    Test getting system information.
    """
    async with DeviceInfo(dvt) as device_info:
        try:
            system_info = await device_info.system_information()
        except UnrecognizedSelectorError:
            pytest.skip("device doesn't support this method")
    assert "_deviceDescription" in system_info and system_info["_deviceDescription"].startswith("Build Version")


@pytest.mark.asyncio
async def test_hardware_information(dvt) -> None:
    """
    Test getting hardware information.
    """
    async with DeviceInfo(dvt) as device_info:
        hardware_info = await device_info.hardware_information()
    assert hardware_info["numberOfCpus"] > 0


@pytest.mark.asyncio
async def test_network_information(dvt) -> None:
    """
    Test getting network information.
    """
    async with DeviceInfo(dvt) as device_info:
        network_info = await device_info.network_information()
    assert network_info["lo0"] == "Loopback"



"""Service-level test fixtures and parametrization for XCUITest integration tests.

XCUITest configurations are loaded from ``tests/services/xcuitest.json``
(not tracked by git).  Copy ``xcuitest.json.example`` to ``xcuitest.json``
and fill in your own bundle identifiers and optional per-config settings.

The file is a JSON array of objects with the following fields:

.. code-block:: json

    [
      {
        "id":              "my_app",
        "runner_bundle_id": "com.example.MyApp.xctrunner",
        "target_bundle_id": "com.example.MyApp",
        "runner_env":      {},
        "timeout":         30.0
      }
    ]

Any test that declares an ``xcuitest_cfg`` parameter is automatically
parametrized over all entries in the JSON file.  If the file is absent the
tests are skipped.

The ``--xcuitest-config`` CLI option allows pointing to a different file:

    pytest --xcuitest-config /path/to/my_config.json ...
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import pytest

_DEFAULT_CONFIG_FILE = Path(__file__).parent / "xcuitest.json"

# Module-level cache: config id -> config dict, populated once per session.
_XCUITEST_CONFIGS: dict[str, dict[str, Any]] = {}


def _load_xcuitest_configs(path: Path) -> list[dict[str, Any]]:
    """Return the list of XCUITest configuration objects from *path*."""
    if not path.exists():
        return []
    return json.loads(path.read_text())


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Parametrize any test that uses the ``xcuitest_cfg`` fixture.

    We parametrize with simple string IDs (not dicts) so that VS Code's pytest
    adapter can serialize the parameter values without crashing.  The actual
    config dict is resolved inside the ``xcuitest_cfg`` fixture below.
    """
    if "xcuitest_cfg" not in metafunc.fixturenames:
        return

    config_path = Path(metafunc.config.getoption("--xcuitest-config") or _DEFAULT_CONFIG_FILE)
    configs = _load_xcuitest_configs(config_path)

    # Populate module-level cache so the fixture can look up by id.
    for cfg in configs:
        _XCUITEST_CONFIGS[cfg["id"]] = cfg

    if not configs:
        metafunc.parametrize(
            "xcuitest_cfg",
            [
                pytest.param(
                    "__skip__",
                    marks=pytest.mark.skip(
                        reason=f"No XCUITest configurations found — create {config_path} "
                        f"(see {config_path.parent / 'xcuitest.json.example'})"
                    ),
                )
            ],
            indirect=True,
        )
    else:
        metafunc.parametrize(
            "xcuitest_cfg",
            [c["id"] for c in configs],
            ids=[c["id"] for c in configs],
            indirect=True,
        )


@pytest.fixture
def xcuitest_cfg(request) -> dict[str, Any]:
    """Resolve the string config ID injected by pytest_generate_tests into the full dict."""
    cfg_id: str = request.param
    return _XCUITEST_CONFIGS[cfg_id]



from pymobiledevice3.services.web_protocol.driver import By
from tests.services.test_web_protocol.common import LINK_HTML


async def test_back(webdriver):
    await webdriver.get("https://www.google.com")
    await webdriver.get("https://github.com")
    await webdriver.back()
    assert (await webdriver.get_current_url()).rstrip("/") == "https://www.google.com"


async def test_current_url(webdriver):
    assert not await webdriver.get_current_url()
    url = "https://www.google.com"
    await webdriver.get(url)
    assert (await webdriver.get_current_url()).rstrip("/") == url


async def test_forward(webdriver):
    await webdriver.get("https://www.google.com")
    await webdriver.get("https://github.com")
    await webdriver.back()
    await webdriver.forward()
    assert (await webdriver.get_current_url()).rstrip("/") == "https://github.com"


async def test_find_element(webdriver):
    await webdriver.execute_script(f"""document.getElementsByTagName('body')[0].innerHTML = '{LINK_HTML}'; """)
    by_id = await webdriver.find_element(By.ID, "id_of_link")
    by_xpath = await webdriver.find_element(By.XPATH, "/html/body/a")
    by_link_text = await webdriver.find_element(By.LINK_TEXT, "the text obviously")
    by_partial_link_text = await webdriver.find_element(By.PARTIAL_LINK_TEXT, "obviously")
    by_name = await webdriver.find_element(By.NAME, "name_of_link")
    by_tag = await webdriver.find_element(By.TAG_NAME, "a")
    by_class = await webdriver.find_element(By.CLASS_NAME, "link-class")
    by_css = await webdriver.find_element(By.CSS_SELECTOR, ".link-class")
    assert by_id == by_xpath == by_link_text == by_partial_link_text == by_name == by_tag == by_class == by_css
    assert by_id is not None



from pathlib import Path

from pymobiledevice3.osu.os_utils import get_os_utils

_OS_UTILS = get_os_utils()
_HOMEFOLDER = _OS_UTILS.get_homedir() / ".pymobiledevice3"


def get_home_folder() -> Path:
    _HOMEFOLDER.mkdir(exist_ok=True, parents=True)
    _OS_UTILS.chown_to_non_sudo_if_needed(_HOMEFOLDER)
    return _HOMEFOLDER



import asyncio
import dataclasses
import json
import logging
import os
import signal
import traceback
import warnings
from contextlib import asynccontextmanager, suppress
from ssl import SSLEOFError
from typing import Optional, Union

import construct

from pymobiledevice3.bonjour import browse_remoted

with warnings.catch_warnings():
    # Ignore: "Core Pydantic V1 functionality isn't compatible with Python 3.14 or greater."
    warnings.simplefilter("ignore", category=UserWarning)
    import fastapi

import uvicorn
from construct import StreamError
from fastapi import FastAPI
from packaging.version import Version

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import (
    ConnectionFailedError,
    ConnectionFailedToUsbmuxdError,
    ConnectionTerminatedError,
    DeviceNotFoundError,
    GetProhibitedError,
    IncorrectModeError,
    InvalidServiceError,
    LockdownError,
    MuxException,
    PairingError,
    QuicProtocolNotSupportedError,
    StreamClosedError,
)
from pymobiledevice3.lockdown import create_using_usbmux, get_mobdev2_lockdowns
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.module_imports import start_tunnel
from pymobiledevice3.remote.remote_service_discovery import RSD_PORT, RemoteServiceDiscoveryService
from pymobiledevice3.remote.tunnel_service import (
    CoreDeviceTunnelProxy,
    RemotePairingProtocol,
    TunnelResult,
    create_core_device_tunnel_service_using_rsd,
    get_remote_pairing_tunnel_services,
)
from pymobiledevice3.remote.utils import get_rsds, stop_remoted
from pymobiledevice3.utils import asyncio_print_traceback

logger = logging.getLogger(__name__)

# bugfix: after the device reboots, it might take some time for remoted to start answering the bonjour queries
REATTEMPT_INTERVAL = 5
REATTEMPT_COUNT = 5

REMOTEPAIRING_INTERVAL = 5
MOBDEV2_INTERVAL = 5

# USB monitor will periodically forget what interfaces it has seen
# and force a full rescan.  The value is number of iterations of the
# inner loop (which sleeps one second each) before blowing away the
# `previous_ips` cache.
USB_MONITOR_RESCAN_INTERVAL = 30

USBMUX_INTERVAL = 2
OSUTILS = get_os_utils()


@dataclasses.dataclass
class TunnelTask:
    task: asyncio.Task
    udid: Optional[str] = None
    tunnel: Optional[TunnelResult] = None


class TunneldCore:
    def __init__(
        self,
        protocol: TunnelProtocol = TunnelProtocol.DEFAULT,
        wifi_monitor: bool = True,
        usb_monitor: bool = True,
        usbmux_monitor: bool = True,
        mobdev2_monitor: bool = True,
    ) -> None:
        self.protocol = protocol
        self.tasks: list[asyncio.Task] = []
        self.tunnel_tasks: dict[str, TunnelTask] = {}
        self.usb_monitor = usb_monitor
        self.wifi_monitor = wifi_monitor
        self.usbmux_monitor = usbmux_monitor
        self.mobdev2_monitor = mobdev2_monitor

    def start(self) -> None:
        """Register all tasks"""
        self.tasks = []
        if self.usb_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_usb_task(), name="monitor-usb-task"))
        if self.wifi_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_wifi_task(), name="monitor-wifi-task"))
        if self.usbmux_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_usbmux_task(), name="monitor-usbmux-task"))
        if self.mobdev2_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_mobdev2_task(), name="monitor-mobdev2-task"))

    def tunnel_exists_for_udid(self, udid: str) -> bool:
        for task in self.tunnel_tasks.values():
            # Linux implementations of `usbmuxd` may report an incorrect value of UDID, dismissing the `-` character.
            # For such cases, we also check for a UDID without it.
            # See: <https://github.com/doronz88/pymobiledevice3/issues/1388#issuecomment-2782249770>
            task_udid = task.udid or ""
            if ((task_udid == udid) or (task_udid.replace("-", "") == udid)) and (task.tunnel is not None):
                return True

        return False

    @asyncio_print_traceback
    async def monitor_usb_task(self) -> None:
        try:
            previous_ips = []
            iteration = 0
            while True:
                iteration += 1
                current_ips = OSUTILS.get_ipv6_ips()
                added = [ip for ip in current_ips if ip not in previous_ips]
                removed = [ip for ip in previous_ips if ip not in current_ips]

                # periodically forget what we have seen so that we reattempt
                # tunnels even if the interface didn't disappear / reappear
                if iteration >= USB_MONITOR_RESCAN_INTERVAL:
                    previous_ips = []
                    iteration = 0
                else:
                    previous_ips = current_ips

                # logger.debug(f'added interfaces: {added}')
                # logger.debug(f'removed interfaces: {removed}')

                for ip in removed:
                    if ip in self.tunnel_tasks:
                        self.tunnel_tasks[ip].task.cancel()
                        with suppress(asyncio.CancelledError):
                            await self.tunnel_tasks[ip].task

                if added:
                    # A new interface was attached
                    for answer in await browse_remoted():
                        for address in answer.addresses:
                            if address.iface.startswith("utun"):
                                # Skip already established tunnels
                                continue
                            if address.full_ip in self.tunnel_tasks:
                                # Skip already established tunnels
                                continue
                            self.tunnel_tasks[address.full_ip] = TunnelTask(
                                task=asyncio.create_task(
                                    self.handle_new_potential_usb_cdc_ncm_interface_task(address.full_ip),
                                    name=f"handle-new-potential-usb-cdc-ncm-interface-task-{address.full_ip}",
                                )
                            )

                # wait before re-iterating
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass

    @asyncio_print_traceback
    async def monitor_wifi_task(self) -> None:
        try:
            while True:
                remote_pairing_tunnel_services = []
                claimed_services = set()
                try:
                    remote_pairing_tunnel_services = await get_remote_pairing_tunnel_services()
                    for service in remote_pairing_tunnel_services:
                        if service.hostname in self.tunnel_tasks:
                            # skip tunnel if already exists for this ip
                            continue
                        if self.tunnel_exists_for_udid(service.remote_identifier):
                            # skip tunnel if already exists for this udid
                            continue
                        self.tunnel_tasks[service.hostname] = TunnelTask(
                            task=asyncio.create_task(
                                self.start_tunnel_task(service.hostname, service),
                                name=f"start-tunnel-task-wifi-{service.hostname}",
                            ),
                            udid=service.remote_identifier,
                        )
                        claimed_services.add(service)
                except asyncio.exceptions.IncompleteReadError:
                    logger.debug("Got IncompleteReadError from monitor-wifi-task")
                except asyncio.CancelledError:
                    # Raise and cancel gracefully
                    raise
                except Exception:
                    logger.error(f"Got exception from {asyncio.current_task().get_name()}")
                finally:
                    for service in remote_pairing_tunnel_services:
                        if service in claimed_services:
                            continue
                        with suppress(Exception):
                            await service.close()
                await asyncio.sleep(REMOTEPAIRING_INTERVAL)
        except asyncio.CancelledError:
            # Cancel gracefully
            pass

    @asyncio_print_traceback
    async def monitor_usbmux_task(self) -> None:
        while True:
            mux = None
            try:
                mux = await usbmux.create_mux()
                await mux.listen()

                while True:
                    await mux.receive_device_state_update()

                    for mux_device in mux.devices:
                        task_identifier = f"usbmux-{mux_device.serial}-{mux_device.connection_type}"
                        if self.tunnel_exists_for_udid(mux_device.serial):
                            # Skip if already established a tunnel for this udid
                            continue
                        if task_identifier in self.tunnel_tasks:
                            # Skip if already trying to establish a tunnel for this device
                            continue
                        service = None
                        try:
                            async with await create_using_usbmux(mux_device.serial) as lockdown:
                                service = await CoreDeviceTunnelProxy.create(lockdown)
                        except (
                            MuxException,
                            InvalidServiceError,
                            GetProhibitedError,
                            construct.core.StreamError,
                            ConnectionTerminatedError,
                            DeviceNotFoundError,
                            LockdownError,
                            IncorrectModeError,
                            SSLEOFError,
                        ):
                            if service is not None:
                                await service.close()
                            continue
                        self.tunnel_tasks[task_identifier] = TunnelTask(
                            udid=mux_device.serial,
                            task=asyncio.create_task(
                                self.start_tunnel_task(task_identifier, service, protocol=TunnelProtocol.TCP),
                                name=f"start-tunnel-task-{task_identifier}",
                            ),
                        )
            except (BlockingIOError, StreamError) as e:
                # Connection lost - will reconnect on next iteration
                logger.debug(f"usbmux connection error: {e}, reconnecting...")
                await asyncio.sleep(USBMUX_INTERVAL)
            except (ConnectionFailedToUsbmuxdError, OSError):
                # This is exception is expected to occur repeatedly on linux running usbmuxd
                # as long as there isn't any physical iDevice connected
                logger.debug("failed to connect to usbmux. waiting for it to restart")
                await asyncio.sleep(USBMUX_INTERVAL)
            except asyncio.CancelledError:
                break
            finally:
                if mux is not None:
                    await mux.close()

    @asyncio_print_traceback
    async def monitor_mobdev2_task(self) -> None:
        try:
            while True:
                async for ip, lockdown in get_mobdev2_lockdowns(only_paired=True):
                    async with lockdown:
                        udid = lockdown.udid
                        task_identifier = f"mobdev2-{udid}-{ip}"
                        if self.tunnel_exists_for_udid(udid):
                            # Skip tunnel if already exists for this udid
                            continue
                        if task_identifier in self.tunnel_tasks:
                            # Skip if already trying to establish a tunnel for this device
                            continue
                        try:
                            tunnel_service = await CoreDeviceTunnelProxy.create(lockdown)
                        except InvalidServiceError:
                            logger.warning(f"[{task_identifier}] failed to start CoreDeviceTunnelProxy - skipping")
                            continue
                    self.tunnel_tasks[task_identifier] = TunnelTask(
                        task=asyncio.create_task(
                            self.start_tunnel_task(task_identifier, tunnel_service),
                            name=f"start-tunnel-task-{task_identifier}",
                        ),
                        udid=udid,
                    )
                await asyncio.sleep(MOBDEV2_INTERVAL)
        except asyncio.CancelledError:
            pass

    @asyncio_print_traceback
    async def start_tunnel_task(
        self,
        task_identifier: str,
        protocol_handler: Union[RemotePairingProtocol, CoreDeviceTunnelProxy],
        queue: Optional[asyncio.Queue] = None,
        protocol: Optional[TunnelProtocol] = None,
    ) -> None:
        if protocol is None:
            protocol = self.protocol
        if isinstance(protocol_handler, CoreDeviceTunnelProxy):
            protocol = TunnelProtocol.TCP
        tun = None
        bailed_out = False
        try:
            if self.tunnel_exists_for_udid(protocol_handler.remote_identifier):
                # cancel current tunnel creation
                raise asyncio.CancelledError()

            async with start_tunnel(protocol_handler, protocol=protocol) as tun:
                if not self.tunnel_exists_for_udid(protocol_handler.remote_identifier):
                    self.tunnel_tasks[task_identifier].tunnel = tun
                    self.tunnel_tasks[task_identifier].udid = protocol_handler.remote_identifier
                    if queue is not None:
                        queue.put_nowait(tun)
                        # avoid sending another message if succeeded
                        queue = None
                    logger.info(f"[{asyncio.current_task().get_name()}] Created tunnel --rsd {tun.address} {tun.port}")
                    await tun.client.wait_closed()
                else:
                    bailed_out = True
                    logger.debug(
                        f"[{asyncio.current_task().get_name()}] Not establishing tunnel since there is already an "
                        f"active one for same udid"
                    )
        except asyncio.CancelledError:
            pass
        except QuicProtocolNotSupportedError as e:
            logger.warning(f"[{asyncio.current_task().get_name()}] {e.__class__.__name__}: {e}")
        except (
            asyncio.exceptions.IncompleteReadError,
            TimeoutError,
            OSError,
            ConnectionResetError,
            StreamError,
            InvalidServiceError,
        ) as e:
            if tun is None:
                logger.debug(f"Got {e.__class__.__name__} from {asyncio.current_task().get_name()}")
            else:
                logger.debug(f"Got {e.__class__.__name__} from tunnel --rsd {tun.address} {tun.port}")
        except Exception:
            logger.exception(f"Got exception from {asyncio.current_task().get_name()}")
        finally:
            if queue is not None:
                # notify something went wrong
                queue.put_nowait(None)

            if tun is not None and not bailed_out:
                logger.info(f"Disconnected from tunnel --rsd {tun.address} {tun.port}")
                await tun.client.stop_tunnel()

            if protocol_handler is not None:
                with suppress(OSError):
                    await protocol_handler.close()

            if task_identifier in self.tunnel_tasks:
                # in case the tunnel was removed just now
                self.tunnel_tasks.pop(task_identifier)

    @asyncio_print_traceback
    async def handle_new_potential_usb_cdc_ncm_interface_task(self, ip: str) -> None:
        rsd = None
        try:
            # establish an untrusted RSD handshake
            rsd = RemoteServiceDiscoveryService((ip, RSD_PORT))

            with stop_remoted():
                first_time = True
                retry = False
                while retry or first_time:
                    retry = False
                    try:
                        await rsd.connect()
                    except StreamClosedError:
                        # Could be on first try because of remoted race
                        if first_time:
                            retry = True
                    except (ConnectionRefusedError, TimeoutError, OSError) as e:
                        raise asyncio.CancelledError() from e
                    finally:
                        first_time = False

            if (self.protocol == TunnelProtocol.QUIC) and (Version(rsd.product_version) < Version("17.0.0")):
                await rsd.close()
                rsd = None
                raise asyncio.CancelledError()

            try:
                core_device_tunnel = await create_core_device_tunnel_service_using_rsd(rsd)
            except InvalidServiceError as e:
                logger.warning(
                    f"[{asyncio.current_task().get_name()}] Skipping {rsd.product_type} ({rsd.product_version}): "
                    f"{e.__class__.__name__}: {e}"
                )
                raise asyncio.CancelledError() from e

            await asyncio.create_task(
                self.start_tunnel_task(ip, core_device_tunnel),
                name=f"start-tunnel-task-usb-{ip}",
            )
        except asyncio.CancelledError:
            pass
        except PairingError:
            logger.exception(f"Failed to pair with {ip}")
        except RuntimeError:
            logger.debug(f"Got RuntimeError from: {asyncio.current_task().get_name()}")
        except Exception:
            logger.exception(f"Error raised from: {asyncio.current_task().get_name()}: {traceback.format_exc()}")
        finally:
            if rsd is not None:
                with suppress(OSError):
                    await rsd.close()

            if ip in self.tunnel_tasks:
                # In case the tunnel was removed just now
                self.tunnel_tasks.pop(ip)

    async def close(self) -> None:
        """close all tasks"""
        for task in self.tasks + [tunnel_task.task for tunnel_task in self.tunnel_tasks.values()]:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    def get_tunnels_ips(self) -> dict:
        """Retrieve the available tunnel tasks and format them as {UDID: [IP]}"""
        tunnels_ips = {}
        for ip, active_tunnel in self.tunnel_tasks.items():
            if (active_tunnel.udid is None) or (active_tunnel.tunnel is None):
                continue
            if active_tunnel.udid not in tunnels_ips:
                tunnels_ips[active_tunnel.udid] = [ip]
            else:
                tunnels_ips[active_tunnel.udid].append(ip)
        return tunnels_ips

    def cancel(self, udid: str) -> None:
        """Cancel active tunnels"""
        for tunnel_ip in self.get_tunnels_ips().get(udid, []):
            self.tunnel_tasks.pop(tunnel_ip).task.cancel()
            logger.info(f"Canceling tunnel {tunnel_ip}")

    def clear(self) -> None:
        """Clear active tunnels"""
        for _udid, tunnel in self.tunnel_tasks.items():
            logger.info(f"Removing tunnel {tunnel}")
            tunnel.task.cancel()
        self.tunnel_tasks = {}


class TunneldRunner:
    """TunneldRunner orchestrate between the webserver and TunneldCore"""

    @classmethod
    def create(
        cls,
        host: str,
        port: int,
        protocol: TunnelProtocol = TunnelProtocol.QUIC,
        usb_monitor: bool = True,
        wifi_monitor: bool = True,
        usbmux_monitor: bool = True,
        mobdev2_monitor: bool = True,
    ) -> None:
        cls(
            host,
            port,
            protocol=protocol,
            usb_monitor=usb_monitor,
            wifi_monitor=wifi_monitor,
            usbmux_monitor=usbmux_monitor,
            mobdev2_monitor=mobdev2_monitor,
        )._run_app()

    def __init__(
        self,
        host: str,
        port: int,
        protocol: TunnelProtocol = TunnelProtocol.QUIC,
        usb_monitor: bool = True,
        wifi_monitor: bool = True,
        usbmux_monitor: bool = True,
        mobdev2_monitor: bool = True,
    ):
        @asynccontextmanager
        async def lifespan(app: FastAPI):
            self._tunneld_core.start()
            yield
            logger.info("Closing tunneld tasks...")
            await self._tunneld_core.close()

        self.host = host
        self.port = port
        self.protocol = protocol
        self._app = FastAPI(lifespan=lifespan)
        self._tunneld_core = TunneldCore(
            protocol=protocol,
            wifi_monitor=wifi_monitor,
            usb_monitor=usb_monitor,
            usbmux_monitor=usbmux_monitor,
            mobdev2_monitor=mobdev2_monitor,
        )

        @self._app.get("/")
        async def list_tunnels() -> dict[str, list[dict]]:
            """Retrieve the available tunnels and format them as {UUID: TUNNEL_ADDRESS}"""
            tunnels = {}
            for ip, active_tunnel in self._tunneld_core.tunnel_tasks.items():
                if (active_tunnel.udid is None) or (active_tunnel.tunnel is None):
                    continue
                if active_tunnel.udid not in tunnels:
                    tunnels[active_tunnel.udid] = []
                tunnels[active_tunnel.udid].append({
                    "tunnel-address": active_tunnel.tunnel.address,
                    "tunnel-port": active_tunnel.tunnel.port,
                    "interface": ip,
                })
            return tunnels

        @self._app.get("/shutdown")
        async def shutdown() -> fastapi.Response:
            """Shutdown Tunneld"""
            os.kill(os.getpid(), signal.SIGINT)
            data = {"operation": "shutdown", "data": True, "message": "Server shutting down..."}
            return generate_http_response(data)

        @self._app.get("/clear_tunnels")
        async def clear_tunnels() -> fastapi.Response:
            self._tunneld_core.clear()
            data = {"operation": "clear_tunnels", "data": True, "message": "Cleared tunnels..."}
            return generate_http_response(data)

        @self._app.get("/cancel")
        async def cancel_tunnel(udid: str) -> fastapi.Response:
            self._tunneld_core.cancel(udid=udid)
            data = {"operation": "cancel", "udid": udid, "data": True, "message": f"tunnel {udid} Canceled ..."}
            return generate_http_response(data)

        @self._app.get("/hello")
        async def hello() -> fastapi.Response:
            data = {"message": "Hello, I'm alive"}
            return generate_http_response(data)

        def generate_http_response(
            data: dict, status_code: int = 200, media_type: str = "application/json"
        ) -> fastapi.Response:
            return fastapi.Response(status_code=status_code, media_type=media_type, content=json.dumps(data))

        @self._app.get("/start-tunnel")
        async def start_tunnel(
            udid: str, ip: Optional[str] = None, connection_type: Optional[str] = None
        ) -> fastapi.Response:
            udid_tunnels = [
                t.tunnel for t in self._tunneld_core.tunnel_tasks.values() if t.udid == udid and t.tunnel is not None
            ]
            if len(udid_tunnels) > 0:
                data = {
                    "interface": udid_tunnels[0].interface,
                    "port": udid_tunnels[0].port,
                    "address": udid_tunnels[0].address,
                }
                return generate_http_response(data)

            queue = asyncio.Queue()
            created_task = False

            try:
                if not created_task and connection_type in ("usbmux", None):
                    task_identifier = f"usbmux-{udid}"
                    try:
                        async with await create_using_usbmux(udid) as lockdown:
                            service = await CoreDeviceTunnelProxy.create(lockdown)
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(
                                task_identifier, service, protocol=TunnelProtocol.TCP, queue=queue
                            ),
                            name=f"start-tunnel-task-{task_identifier}",
                        )
                        self._tunneld_core.tunnel_tasks[task_identifier] = TunnelTask(task=task, udid=udid)
                        created_task = True
                    except (ConnectionFailedError, InvalidServiceError, MuxException):
                        pass
                if connection_type in ("usb", None):
                    for rsd in await get_rsds(udid=udid):
                        rsd_ip = rsd.service.address[0]
                        if ip is not None and rsd_ip != ip:
                            await rsd.close()
                            continue
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(
                                rsd_ip, await create_core_device_tunnel_service_using_rsd(rsd), queue=queue
                            ),
                            name=f"start-tunnel-usb-{rsd_ip}",
                        )
                        self._tunneld_core.tunnel_tasks[rsd_ip] = TunnelTask(task=task, udid=rsd.udid)
                        created_task = True
                if not created_task and connection_type in ("wifi", None):
                    for remotepairing in await get_remote_pairing_tunnel_services(udid=udid):
                        remotepairing_ip = remotepairing.hostname
                        if ip is not None and remotepairing_ip != ip:
                            await remotepairing.close()
                            continue
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(remotepairing_ip, remotepairing, queue=queue),
                            name=f"start-tunnel-wifi-{remotepairing_ip}",
                        )
                        self._tunneld_core.tunnel_tasks[remotepairing_ip] = TunnelTask(
                            task=task, udid=remotepairing.remote_identifier
                        )
                        created_task = True
            except Exception as e:
                return fastapi.Response(
                    status_code=501,
                    content=json.dumps({
                        "error": {
                            "exception": e.__class__.__name__,
                            "traceback": traceback.format_exc(),
                        }
                    }),
                )

            if not created_task:
                return fastapi.Response(status_code=501, content=json.dumps({"error": "task not created"}))

            tunnel: Optional[TunnelResult] = await queue.get()
            if tunnel is not None:
                data = {"interface": tunnel.interface, "port": tunnel.port, "address": tunnel.address}
                return generate_http_response(data)
            else:
                return fastapi.Response(
                    status_code=404, content=json.dumps({"error": "something went wrong during tunnel creation"})
                )

    def _run_app(self) -> None:
        uvicorn.run(self._app, host=self.host, port=self.port, loop="asyncio")
